# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import concurrent.futures
from collections.abc import Callable
from typing import Any


# Sentinels/Signals
class OptionSucceeded(Exception):
    pass


class FailedParse(Exception):
    pass


class ChoiceContext:
    def __init__(self, ctx: Any, parallel: bool = False):
        self.ctx = ctx
        self.parallel: bool = parallel
        self.options: list[Callable[[Any], None]] = []
        self.expected: list[str] = []

    def option(self, func: Callable[[Any], None]) -> Callable[[Any], None]:
        """Expression to register a grammar branch."""
        self.options.append(func)
        return func

    def expecting(self, *tokens: str) -> None:
        """Register tokens for error reporting if all options fail."""
        self.expected.extend(tokens)

    @staticmethod
    def _worker(ctx: Any, opt: Callable[[Any], None], ss: Any) -> Any:
        """
        Isolated execution for a single option.
        Returns the mutated 'ss' on success, or the Exception on failure.
        """
        try:
            with ctx._option(ss):
                opt(ss)
            return ss
        except Exception as e:
            return e

    def run(self, ss: Any) -> None:
        """
        Orchestrates the execution of registered options.
        """
        if not self.options:
            return

        if self.parallel:
            self._run_parallel(ss)
        else:
            self._run_sequential(ss)

        # If we reach here, no options succeeded.
        raise self.ctx.newexcept(f"Expected one of: {', '.join(self.expected)}")

    def _run_sequential(self, ss: Any) -> None:
        """Standard PEG backtracking: one option at a time."""
        for opt in self.options:
            try:
                with self.ctx._option(ss):
                    opt(ss)
                raise OptionSucceeded()
            except FailedParse:
                continue

    def _run_parallel(self, ss: Any) -> None:
        """
        Concurrent execution using the shared engine executor.
        Maintains PEG priority by resolving futures in registration order.
        """
        executor: concurrent.futures.ThreadPoolExecutor = self.ctx.executor
        futures: list[concurrent.futures.Future] = []

        try:
            # 1. Dispatch branches
            for opt in self.options:
                futures.append(
                    executor.submit(self._worker, self.ctx, opt, ss.branch())
                )

            # 2. Resolve in order to preserve PEG priority
            for future in futures:
                try:
                    result = future.result()

                    # If the worker caught an exception, raise it here
                    if isinstance(result, Exception):
                        raise result

                    # 3. Success: Merge worker state and signal completion
                    self.ctx.merge_state(ss, result)
                    raise OptionSucceeded()

                except FailedParse:
                    # Expected failure: move to the next future in registration order
                    continue
                except (OptionSucceeded, Exception):
                    # Success or Fatal Error: Clean up and propagate
                    self._cancel_all(futures)
                    raise

        finally:
            self._cancel_all(futures)

    def _cancel_all(self, futures: list[concurrent.futures.Future]) -> None:
        """Cancels all pending futures in the current choice set."""
        for f in futures:
            if not f.done():
                f.cancel()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import pytest

import tatsu
from tatsu.exceptions import FailedParse


class TestFailedParseRender:
    def test_render_features_present(self):
        grammar = r'''
            @@grammar :: Test

            start = 'hello' 'world' ;
        '''
        model = tatsu.compile(grammar=grammar)

        with pytest.raises(FailedParse) as exc_info:
            model.parse('hello missing', source="example")

        e = exc_info.value
        r = e.render()
        # print(r)
        # assert False

        assert 'error' in r
        assert '->' in r
        assert '│' in r
        assert 'start' in r
        assert isinstance(r, str)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause


def forward[T](subject: T) -> T:
    return subject


@forward
def f() -> int: ...  # pyright: ignore[reportRedeclaration]  # ty:ignore[empty-body]


assert f() is None


def f() -> int:
    return 0


assert f() == 0



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations


"""Character-based progress bar visual using fill/empty markers."""

from ..util import findfirst
from ..ztyle import Style


__all__ = ["Bar"]


class Bar:
    def __init__(
        self,
        /,
        total: int = 1,
        *,
        width: int = 0,
        fill: str = "=>.",  # pyright: ignore[reportRedeclaration]
        style: list[Style] | None = None,
    ):
        self.pos: int = 0
        self.total: int = total
        self.width: int = width
        self.fill: str = (fill + "...")[:3]

        self.style: list[Style] = style or []
        s = Style()
        if not self.style:
            self.style = [s] * 3
        elif len(self.style) == 2:
            self.style = [*self.style, s]
        else:
            self.style = [self.style[0]] * 2 + [s]

    def __str__(self) -> str:
        return self.render(max(1, self.total))

    def __len__(self) -> int:
        return self.width

    def __format__(self, fmt: str) -> str:
        ws = findfirst(r"(\d+).?", fmt)
        w = int(ws) if ws else self.total
        return self.render(w)

    def update(self, pos: int, total: int, fill: str | None = None) -> None:
        self.pos = pos
        self.total = total
        if fill is not None:
            self.fill = fill

    def render(self, budget: int) -> str:
        total = max(1, self.pos, self.total)

        pos = self.pos
        pos = max(0, min(total, self.pos))

        done_w = int(pos / total * budget)
        todo_w = budget - done_w

        dones = self.render_done(done_w)
        todos = self.render_todo(todo_w)

        return self.trim_to_width(budget, f"{dones}{todos}")

    def render_done(self, w: int) -> str:
        done = self.fill[0]
        sd = self.style[0]
        if self.pos >= self.total:
            return sd.apply(done * w)

        head = self.fill[1]
        sh = self.style[1]
        if sh == sd:
            return sd.apply(done * (w - 1) + head)
        return sd.apply(done * (w - 1)) + sh.apply(head)

    def render_todo(self, w: int) -> str:
        todo = self.fill[2]
        st = self.style[2]
        return st.apply(todo * w)

    def trim_to_width(self, budget: int, rendered: str) -> str:
        from ..ztyle import visual_len as vlen

        while (w := vlen(rendered)) > budget:
            chars = [
                self.fill[1],
                self.fill[2],
                self.fill[0],
                "█",
                "░",
                ">",
                "-",
                "=",
                ".",
            ]
            for char in chars:
                rendered = rendered.replace(char, "", 1)
                if vlen(rendered) < w:
                    break
        return rendered



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import builtins
import types
from collections.abc import Callable, Mapping
from dataclasses import dataclass, field
from typing import Any, cast

from ..util import (
    Config,
    Constructor,
    TypeContainer,
    boundcall,
    deprecated_params,
    fqn,
    fqntype,
    least_upper_bound_type,
    mangle,
)
from .node import Node
from .synth import synthesize


class TypeResolutionError(TypeError):
    """Raised when a constructor for a node type cannot be found or synthesized"""

    pass


def types_defined_in(container: TypeContainer) -> list[type]:
    return ModelBuilder.types_defined_in(container)


@dataclass
class BuilderConfig(Config):
    basetype: type = Node
    synthok: bool = True
    typedefs: list = field(default_factory=list)
    constructors: list = field(default_factory=list)

    def __getstate__(self) -> dict[str, Any]:
        state = super().__getstate__()

        state['typedefs'] = [fqn(t) for t in self.typedefs]
        state['constructors'] = [fqn(c) for c in self.constructors]

        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        super().__setstate__(state)

        self.typedefs = [fqntype(t) for t in state.get('typedefs', [])]
        self.constructors = [fqntype(c) for c in state.get('constructors', [])]


class ModelBuilder:
    """Intended as a semantic action for parsing. A ModelBuildercreates
    nodes using the class name given as first parameter to a grammar
    rule, and synthesizes the class/type if it's not known.
    """

    def __init__(
        self,
        config: BuilderConfig | None = None,
        synthok: bool = True,
        basetype: type | None = None,
        typedefs: list[TypeContainer] | None = None,
        constructors: list[Constructor] | None = None,
    ) -> None:
        config = BuilderConfig.new(
            config=config,
            basetype=basetype,
            synthok=synthok,
            typedefs=typedefs,
            constructors=constructors,
        )
        self._registry: dict[str, Constructor] = {}

        config = self._typedefs_to_constructors(config)
        config = self._narrow_basetype(config)
        self._register_constructors(config)

        self.config = config
        # HACK!
        self.config = self.config.override(synthok=not constructors)

    def instanceof(
        self,
        typename,
        /,
        known: dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        return self._instanceof(typename, known, *args, **kwargs)

    def _instanceof(
        self,
        typename: str,
        known: dict[str, Any],
        /,
        *args: Any,
        base: type | None = None,
        **kwargs: Any,
    ) -> Any:
        constructor = self._get_constructor(typename, base=base)
        return boundcall(constructor, known, *args, **kwargs)

    def safe_context(self, *other: Mapping[str, Any]) -> Mapping[str, Any]:
        context: dict[str, Any] = {}
        for ctx in other:
            context |= ctx

        return {name: value for name, value in context.items() if callable(value)}

    @staticmethod
    def types_defined_in(container: TypeContainer, /) -> list[type]:
        contents: dict[str, Any] = {}
        if isinstance(container, types.ModuleType):
            contents.update(vars(container))
        elif isinstance(container, Mapping):
            contents.update(container)  # ty:ignore[no-matching-overload]
        elif hasattr(container, '__dict__'):
            contents.update(vars(container))
        else:
            return []

        type_list = [
            typ
            for name, typ in contents.items()
            if not name.startswith('_') and isinstance(typ, type)
        ]
        name = getattr(container, '__module__', None) or getattr(
            container,
            '__name__',
            None,
        )
        if name is None:
            return type_list
        return [t for t in type_list if t.__module__ == name]

    def _funname(self, obj: Any) -> str | None:
        name = getattr(obj, '__name__', None)
        if not name and callable(obj):
            name = type(obj).__name__
        return name

    def _typedefs_to_constructors(self, config: BuilderConfig) -> BuilderConfig:
        if not config.typedefs:
            return config

        contained = []
        for container in config.typedefs:
            contained += self.types_defined_in(container)

        all_constructors = [*config.constructors, *contained]
        return config.override(constructors=all_constructors)

    def _narrow_basetype(self, config: BuilderConfig) -> BuilderConfig:
        if config.basetype and config.basetype is not object:
            return config
        basetype = least_upper_bound_type(config.constructors)
        return config.override(basetype=basetype)

    def _register_constructors(self, config: BuilderConfig) -> None:
        for t in config.constructors:
            if not callable(t):
                raise TypeResolutionError(
                    f'Expected callable in constructors, got: {type(t)!r}',
                )

            name = self._funname(t)
            if not name:
                raise TypeResolutionError(
                    f'Expected __name__ in constructor, got: {type(t)!r}',
                )

            # NOTE: this allows for standalone functions as constructors
            self._register_constructor(t)

    def _register_constructor(self, constructor: Constructor) -> Constructor:
        name = self._funname(constructor)
        if not name:
            return constructor

        existing = self._registry.get(name)
        if existing and existing is not constructor:
            raise TypeResolutionError(
                f"Conflict for constructor name {name!r}:"
                f" attempted to register {constructor!r},"
                f" but {existing!r} is already registered"
                f"{id(existing)=} {id(constructor)=}."
            )

        self._registry[name] = constructor
        return constructor

    def _find_existing_constructor(self, typename: str) -> Constructor | None:
        return self._registry.get(typename) or vars(builtins).get(typename)

    def _get_constructor(
        self,
        typename: str,
        base: type | None,
        **args: Any,
    ) -> Constructor:
        if constructor := self._find_existing_constructor(typename):
            return constructor

        if not self.config.synthok:
            synthok = bool(self.config.synthok)
            raise TypeResolutionError(
                f'Could not find constructor for type {typename!r}, and {synthok=} ',
            )

        if base is None:
            constructor = synthesize(typename, (), **args)
        else:
            constructor = synthesize(typename, (base,), **args)

        return self._register_constructor(constructor)

    def __getstate__(self) -> dict[str, Any]:
        state: dict[str, Any] = cast(dict, super().__getstate__())
        state['__registry'] = [fqn(t) for t in self._registry]
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        state['_registry'] = [fqntype(t) for t in state.get('_registrt', [])]
        for name, value in state.items():
            setattr(self, name, value)


class ModelBuilderSemantics:
    @deprecated_params(base_type='basetype', types='constructors', context=None)
    def __init__(
        self,
        config: BuilderConfig | None = None,
        synthok: bool = True,
        basetype: type | None = None,
        typedefs: list[TypeContainer] | None = None,
        constructors: list[Constructor] | None = None,
        # deprecations
        context: Any | None = None,
        base_type: type | None = None,
        types: list[Callable] | None = None,  # for bw compatibility
    ) -> None:
        config = BuilderConfig.new(
            config=config,
            basetype=basetype,
            synthok=synthok,
            typedefs=typedefs,
            constructors=constructors,
        )
        # handle deeprecations
        assert context is None
        if isinstance(base_type, type):
            config = config.override(basetype=base_type)
        if types is not None:
            constructors = [*(config.constructors or []), *types]
            config = config.override(constructors=constructors)

        self.config = config
        self._builder = ModelBuilder(
            config=config,
            synthok=synthok,
            basetype=basetype,
            typedefs=typedefs,
            constructors=constructors,
        )

    @property
    def builder(self) -> ModelBuilder:
        return self._builder

    @staticmethod
    def types_defined_in(container: TypeContainer, /) -> list[type]:
        return ModelBuilder.types_defined_in(container)

    def _default(self, ast: Any, *args: Any, **kwargs: Any) -> Any:
        if not args:
            return ast

        typespec = [mangle(s) for s in args[0].split('::')]
        typename = typespec[0]
        basenames = reversed(typespec)

        base = self.config.basetype
        for basename in basenames:
            defined = self._builder._get_constructor(basename, base=base)
            if isinstance(defined, type):
                base = defined

        known = {'ast': ast, 'exp': ast}
        return self._builder._instanceof(
            typename,
            known,
            ast,
            *args[1:],
            base=base,
            **kwargs,
        )



import base64
import textwrap
import zlib


def compress_to_ascii(text: str, width: int = 80) -> str:
    """Compresses a string and encodes it into wrapped printable ASCII."""
    # 1. Convert text to raw UTF-8 bytes
    raw_bytes = text.encode("utf-8")

    # 2. Compress the bytes using zlib (DEFLATE)
    compressed_bytes = zlib.compress(raw_bytes, level=9)

    # 3. Encode binary payload into clean, printable Base85 ASCII characters
    ascii_bytes = base64.b85encode(compressed_bytes)
    ascii_str = ascii_bytes.decode("ascii")

    # 4. Wrap beautifully to the target column width
    return "\n".join(textwrap.wrap(ascii_str, width=width))


def decompress_from_ascii(ascii_block: str) -> str:
    """Reverses the pipeline: unwraps, decodes Base85, and decompresses."""
    # 1. Remove all whitespace and newlines introduced by column wrapping
    cleaned_ascii = "".join(ascii_block.split())

    # 2. Decode the Base85 string back to binary zlib payload
    compressed_bytes = base64.b85decode(cleaned_ascii.encode("ascii"))

    # 3. Decompress the data structure back to standard UTF-8 text
    raw_bytes = zlib.decompress(compressed_bytes)
    return raw_bytes.decode("utf-8")



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from ..util.fromjson import JSONBase


"""Computed metrics and column resolution for a progress bar row."""

import datetime as dt
from functools import cached_property
from typing import Any, Protocol

from ..util import clock_time_μs
from ..ztyle import Style
from .bar import Bar
from .col import Col


__all__ = ["BarRowProtocol", "Metrics"]


class BarRowProtocol(Protocol):
    @property
    def id(self) -> str: ...

    @property
    def state(self) -> int: ...

    @property
    def begun(self) -> int: ...

    @property
    def pos(self) -> int: ...

    @property
    def total(self) -> int: ...

    @property
    def width(self) -> int: ...

    @property
    def label(self) -> str | Style: ...

    @property
    def cols(self) -> list[Any]: ...

    @property
    def style(self) -> list[Style]: ...

    @property
    def fill(self) -> str: ...

    @property
    def selfstop(self) -> bool: ...


class Metrics(JSONBase):
    """Lazy, cached metrics for a BarRow. Create fresh per render cycle."""

    def __init__(self, row: BarRowProtocol) -> None:
        self.row = row

    def resolve(self, col: Col) -> Any:
        return getattr(self, col.value)

    @property
    def id(self) -> str:
        return self.row.id

    @property
    def state(self) -> int:
        return self.row.state

    @property
    def label(self) -> str | Style:
        return self.row.label

    @property
    def padding(self) -> str:
        return " "

    @property
    def pos(self) -> int:
        return self.row.pos

    @property
    def total(self) -> int:
        return self.row.total

    @property
    def fill(self) -> str:
        return self.row.fill

    @property
    def style(self) -> list[Style]:
        return self.row.style or []

    @property
    def width(self) -> int:
        return self.row.width

    @property
    def begun(self) -> int:
        return self.row.begun

    # -- computed, cached lazily --
    @cached_property
    def elapsed(self) -> int:
        return max(0, clock_time_μs() - self.begun)

    @cached_property
    def rt(self) -> dt.timedelta:
        return dt.timedelta(seconds=self.elapsed / 1_000_000_000)

    @cached_property
    def pct(self) -> float:
        return self.pos / self.total if self.total else 0.0

    @cached_property
    def percentage(self) -> float:
        return 100 * self.pct

    @cached_property
    def eta_s(self) -> float:
        if not self.pct:
            return 0.0
        return max(0.0, self.elapsed / self.pct - self.elapsed)

    @cached_property
    def eta(self) -> dt.timedelta:
        return dt.timedelta(seconds=self.eta_s)

    @cached_property
    def h(self) -> int:
        return self.elapsed // 3_600_000_000_000

    @cached_property
    def m(self) -> int:
        return (self.elapsed // 60_000_000_000) % 60

    @cached_property
    def s(self) -> int:
        return (self.elapsed // 1_000_000_000) % 60

    @cached_property
    def ms(self) -> int:
        return (self.elapsed // 1_000_000) % 1000

    @cached_property
    def bar(self):
        b = Bar(
            fill=self.fill,
            style=self.style,
            width=self.width,
        )
        b.update(self.pos, self.total)
        return b



# Adapted from pegen (sccutils.py) under the MIT license.
# pegen is Copyright (c) 2019, Guido van Rossum and contributors.

# noqa # type: ignore # ruff: noqa
# ty: ignore-file
# pyrefly: ignore
# pyright: ignore
# pyright: reportGeneralTypeIssues=false

from typing import AbstractSet, Dict, Iterable, Iterator, List, Set


def strongly_connected_components(
    vertices: AbstractSet[str], edges: Dict[str, AbstractSet[str]]
) -> Iterator[AbstractSet[str]]:
    identified: Set[str] = set()
    stack: List[str] = []
    index: Dict[str, int] = {}
    boundaries: List[int] = []

    def dfs(v: str) -> Iterator[Set[str]]:
        index[v] = len(stack)
        stack.append(v)
        boundaries.append(index[v])

        for w in edges[v]:
            if w not in index:
                yield from dfs(w)
            elif w not in identified:
                while index[w] < boundaries[-1]:
                    boundaries.pop()

        if boundaries[-1] == index[v]:
            boundaries.pop()
            scc = set(stack[index[v] :])
            del stack[index[v] :]
            identified.update(scc)
            yield scc

    for v in vertices:
        if v not in index:
            yield from dfs(v)


def find_cycles_in_scc(
    graph: Dict[str, AbstractSet[str]], scc: AbstractSet[str], start: str
) -> Iterable[List[str]]:
    assert start in scc, (start, scc)
    assert scc <= graph.keys(), scc - graph.keys()

    graph = {
        src: {dst for dst in dsts if dst in scc}
        for src, dsts in graph.items()
        if src in scc
    }
    assert start in graph

    def dfs(node: str, path: List[str]) -> Iterator[List[str]]:
        if node in path:
            yield path + [node]
            return
        path = path + [node]
        for child in graph[node]:
            yield from dfs(child, path)

    yield from dfs(start, [])



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# from: https://www.unicode.org/charts/nameslist/n_2500.html
from __future__ import annotations

import pytest

from tatsu.util import countlines, lcnt, linecount, strtools


def test_visible_width():
    assert strtools.unicode_display_len("abc") == 3
    assert strtools.unicode_display_len("Python") == 6
    assert strtools.unicode_display_len("蛇") == 2
    assert strtools.unicode_display_len("🐍 Py") == 5


@pytest.mark.parametrize(
    "text, expected",
    [
        ("", 1),
        ("hello", 1),
        ("hello\n", 2),
        ("\n\n", 3),
        ("win\r\nline", 2),
        ("mac\rline", 2),
        ("n\ro\r\nw", 3),
        ("line1\nline2", 2),
    ],
)
def test_visual_linecount(text, expected):
    assert linecount(text) == expected


@pytest.mark.parametrize(
    "text, delta",
    [
        ("", 1),
        ("hello", 0),
        ("line1\nline2", 0),
        ("hello\n", 1),
        ("\n\n", 1),
        ("n\ro\r\nw", 0),
        ("n\ro\r\nw\n", 1),
        ("win\r\n", 1),
        ("mac\r", 1),
    ],
)
def test_linecount_delta(text, delta):
    assert linecount(text) - delta == len(text.splitlines())


@pytest.mark.parametrize(
    "text, expected",
    [
        ("", False),
        ("hello", False),
        ("hello\n", True),
        ("\n\n", True),
        ("win\r\nline", True),
        ("mac\rline", True),
        ("n\ro\r\nw", True),
        ("line1\nline2", True),
    ],
)
def test_ismultiline(text, expected):
    assert strtools.ismultiline(text) == expected


@pytest.mark.parametrize(
    "text, expect, lcount",
    [
        ("", lcnt(0, 0, 0, 0), 0),
        ("x=1", lcnt(1, 0, 0, 1), 1),
        ("\n", lcnt(1, 1, 0, 0), 1),
        ("\r\n\r\n", lcnt(2, 2, 0, 0), 2),
        ("x=1\n", lcnt(1, 0, 0, 1), 1),
        ("# comment", lcnt(1, 0, 1, 0), 1),
        ("  # indented", lcnt(1, 0, 1, 0), 1),
        ("x=1 # inline", lcnt(1, 0, 0, 1), 1),
        (" \t \n#\ncode", lcnt(3, 1, 1, 1), 3),
        ("x=1\n#\n\n#\ny=2", lcnt(5, 1, 2, 2), 5),
    ],
)
def test_sloc_consistency(text, expect, lcount):
    actual = countlines(text)

    # assert [text, lcount] == [text, linecount(text)]
    assert [text, actual.totl] == [text, lcount]
    assert [text, actual.totl] == [text, len(text.splitlines())]

    assert [text, actual] == [text, expect]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import types


def is_rich_library_available() -> bool:
    try:
        import rich  # noqa: F401  # pyright: ignore[reportMissingImports]

        return True
    except ImportError:
        return False


def get_rich() -> types.ModuleType:
    import rich  # noqa: F401  # pyright: ignore[reportMissingImports]

    return rich



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import sys

from .cli import cling_main


if __name__ == "__main__":
    sys.exit(cling_main())



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re


# Compiles standard 7-bit ANSI control sequences (CSI, SGR, etc.)
ANSI_RE = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
SGR_RE = re.compile(r"\x1B\[([\d;]*)m")


def tty_escape(s: str) -> str:
    return s.replace('\x1b', '\\e').replace('\\x1b', '\\e')


def tty_unescape(s: str) -> str:
    return s.replace('\\e', '\x1b')


def descape(text: str | bytes) -> str:
    """Removes ANSI escape codes from a string."""
    if isinstance(text, bytes):
        text = str(text)
    if not isinstance(text, str):
        raise TypeError(f"expected str got {type(text)!r}")
    return ANSI_RE.sub("", text)


def visual_len(text: str | bytes) -> int:
    """Returns the true visual length of a string, omitting
    terminal escape codes."""
    return len(descape(text))


def hide_cursor() -> str:
    """Returns an escape sequence that hides the cursor."""
    return "\033[?25l"


def show_cursor() -> str:
    """Returns an escape sequence that shows the cursor."""
    return "\033[?25h"


def clearline(text: str = "") -> str:
    """Returns the text prefixed with a clear-to-end-of-line escape sequence."""
    return f"\033[K{text}\n"


def shoot_lines(texts: list[str]) -> list[str]:
    """Returns a block of clear-to-end-of-line escape sequences for each line."""
    return [clearline(t) for t in texts]


def pushup(lines: int) -> str:
    """Returns an escape sequence that moves the cursor up a given number of lines."""
    return f"\033[{lines}A"


def blankpad(count: int) -> list[str]:
    """Generates a block of empty, cleared lines to wipe out stale terminal trailing rows."""
    return ["\033[K\n"] * count



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import multiprocessing
import threading
from collections.abc import Iterable
from typing import Any, Protocol

from ..util import identity
from .pmap import HAS_MULTITHREADING_SUPPORT, active_pmap
from .result import Result
from .task import Event, Func, Task, taskproc


__all__ = [
    'Progress',
    'parproc',
]


class Progress(Protocol):
    def update(self, *args, **kwargs) -> None: ...
    def stop(self) -> None: ...


def parproc(
    func: Func,
    payloads: Iterable[Any],
    /,
    *args: Any,
    pickable: Func = identity,
    parallel: bool = True,
    reraise: bool = False,
    max_workers: int | None = None,
    **kwargs: Any,
) -> Iterable[Result]:
    stop: Event = threading.Event()
    if not HAS_MULTITHREADING_SUPPORT:
        stop = multiprocessing.Manager().Event()

    tasks = [
        Task(
            stop=stop,
            func=func,
            payload=payload,
            pickable=pickable,
            reraise=reraise,
            args=args,
            kwargs=kwargs,
        )
        for payload in payloads
    ]

    if len(tasks) == 1:
        yield taskproc(tasks[0])
        return

    if not parallel:
        yield from map(taskproc, tasks)
    else:
        pmap = active_pmap()
        yield from pmap(stop, taskproc, tasks, max_workers)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import importlib
import importlib.util
from pathlib import Path

from .typetools import cast


def importall(modulename: str) -> dict[str, object]:
    """
    Simulate 'from modulename import *' and return the resulting attributes as a dict.
    """
    module = importlib.import_module(modulename)

    if hasattr(module, "__all__"):
        keys: list[str] = cast(list[str], module.__all__)  # pyright: ignore[reportAny]
    else:
        keys = [n for n in dir(module) if not n.startswith("_")]

    return {k: getattr(module, k) for k in keys}


def module_available(name: str) -> bool:
    return importlib.util.find_spec(name) is not None


def module_missing(name: str) -> bool:
    return not module_available(name)


def pathtomodulename(path: str | Path) -> str:
    path = Path(path) if isinstance(path, str) else path
    return str(path.with_suffix('')).replace('/', '.').replace('.__init__', '')



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import Any

from .cfg import CLIConfig, CLIError
from .fmt import colorize_output


def output_results(cfg: CLIConfig, results: list[tuple[str, Any]]) -> None:
    out_path = Path(cfg.output) if cfg.output else None
    if not results:
        return

    # Directory output: each input gets its own .json file
    if out_path and (
        str(cfg.output).rstrip().endswith((os.sep, "/")) or out_path.is_dir()
    ):
        out_path.mkdir(parents=True, exist_ok=True)
        ext = _output_ext(cfg)
        for input_path, single_payload in results:
            name = Path(input_path).stem
            out = (out_path / name).with_suffix(ext)
            _show(single_payload, out)
        return

    # JSONL for multiple input files (JSON is the default format)
    # Skip JSONL when output is to a TTY — use indented JSON instead.
    if cfg.json_lines or (
        (not cfg.model or cfg.json)
        and len(results) > 1
        and not (not cfg.output and sys.stdout.isatty())
    ):
        import json

        jsonl = "\n".join(
            json.dumps(
                {"input": input_path, "result": json.loads(outcome)},
                separators=(",", ":"),
            )
            for input_path, outcome in results
        )
        _show(jsonl, _output_path(cfg))
        return

    # Single result or model output: write sequentially
    single_out = _output_path(cfg)
    should_colorize = (
        not cfg.json_lines and not cfg.railroads and cfg.usecolor and single_out is None
    )

    language = "json"
    if cfg.model:
        language = "python"
    elif cfg.pretty:
        language = "ebnf"

    payloads = [payload for _, payload in results]
    if should_colorize:
        payloads = [
            colorize_output(payload, language, cfg.style) for payload in payloads
        ]

    single_payload = "\n".join(payloads)
    if single_out is None:
        print(single_payload)
    else:
        _make_output_dir(single_out).write_text(single_payload)


def _output_ext(cfg: CLIConfig) -> str:
    """Return file extension for the current output format."""
    if cfg.model:
        return ".py"
    if cfg.railroads:
        return ".railroads.txt"
    if cfg.pretty:
        return ".ebnf"
    return ".json"


def _output_path(cfg: CLIConfig, *, name: str | None = None) -> Path | None:
    """Return output path, or None for stdout.

    If cfg.output is a directory, appends *name* (with format extension).
    Otherwise returns the file path.
    """
    if not cfg.output:
        return None
    out = Path(cfg.output)
    if out.is_dir():
        if name is None:
            return None
        return out / (name + _output_ext(cfg))
    return out


def _make_output_dir(path: Path | str) -> Path:
    """Ensure the output directory exists and return the path."""
    path = Path(path)
    dir = path.parent
    if dir.exists() and not dir.is_dir():
        raise CLIError(f"Output path exists but it's not a directory: {dir!s}")
    try:
        dir.mkdir(parents=True, exist_ok=True)
    except OSError as e:
        raise CLIError(f"Failed to create output directory: {e!s}") from e
    return path


def _show(payload: str, output: Path | None) -> None:
    """Write payload to a file or stdout."""
    if output is None:
        print(payload)
    else:
        _make_output_dir(output).write_text(payload)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from ..api import boot_grammar
from .cfg import CLIConfig
from .fmt import render_grammar
from .global_opt import add_global_options
from .grammar_cmd import add_grammar_options
from .lib import Results


def boot_cmd(cfg: CLIConfig) -> Results:
    """Handle the ``boot`` subcommand."""

    payload = render_grammar(
        boot_grammar(),
        cfg,
        name="boot",
    )
    return [("boot", payload)]


def add_boot_cmd(subparsers):
    boot_parser = subparsers.add_parser(
        "boot",
        help="The internal boot grammar",
    )
    add_global_options(boot_parser)
    add_grammar_options(boot_parser)
    return boot_parser



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
# NOTE: re-exports mostly for bw compatibility
from __future__ import annotations

import sys

from .abctools import *  # noqa: F403
from .asjson import *  # noqa: F403
from .common import *  # noqa: F403
from .configs import *  # noqa: F403
from .debugging import *  # noqa: F403
from .deprecate import *  # noqa: F403
from .genericmain import generic_main as generic_main  # noqa: F403
from .indent import *  # noqa: F403
from .itertools import *  # noqa: F403
from .misc import *  # noqa: F403
from .moduletools import *  # noqa: F403
from .regextools import *  # noqa: F403
from .safeeval import *  # noqa: F403
from .strtools import *  # noqa: F403
from .typetools import *  # noqa: F403
from .undefined import *  # noqa: F403
from .version import *  # noqa: F403


def _legacy_parproc() -> None:  # noqa
    from .. import parproc

    sys.modules['tatsu.util.parproc'] = parproc


_legacy_parproc()  # noqa



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from dataclasses import dataclass
from threading import Event
from typing import Any

from ..packetz import Packet


@dataclass(slots=True, order=True)
class Result(Packet):
    stop: Event
    payload: Any
    outcome: Any = None
    exception: Any = None
    linecount: int = 0
    runtime: float = 0
    memory: int = 0

    @property
    def success(self):
        return self.exception is None

    def __str__(self):
        return str(self.__dict__)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Self, overload

from .ast import AST
from .cst import cstadd, cstfinal, cstmerge
from .infos import Alert, RuleInfo


__all__ = ['ParseState', 'ParseStateStack']

from ..input import Cursor


_AT_ = '__vallue__'


class ParseState:
    __slots__ = (
        'alerts',
        'ast',
        'cst',
        'cursor',
        'cutseen',
        'last_node',
    )

    def __init__(self, base: Cursor | ParseState) -> None:
        if isinstance(base, ParseState):
            self.cursor: Cursor = base.cursor.clone()
            self.ast: AST = AST(base.ast)
        else:
            self.cursor = base.clone()
            self.ast = AST()

        self.cst: Any = None
        self.cutseen: bool = False
        self.last_node: Any = None
        self.alerts: list[Alert] = []

    def clone(self) -> Self:
        return type(self)(self)

    def __copy__(self) -> Self:
        new = type(self)(self.cursor)
        new.alerts = self.alerts[:]
        return new

    def merge(self, prev: ParseState) -> Self:
        self.ast = prev.ast
        self.extend(prev.cst)
        self.alerts.extend(prev.alerts)
        self.cursor.goto(prev.cursor.pos)
        return self

    def fold(self) -> Any:
        ast = self.ast
        cst = self.cst
        if not ast:
            return cstfinal(cst)
        elif _AT_ in ast:
            return ast[_AT_]
        else:
            return ast

    def append(self, node: Any) -> Any:
        self.last_node = node
        self.cst = cstadd(self.cst, node)
        return node

    def extend(self, node: Any) -> Any:
        self.last_node = node
        self.cst = cstmerge(self.cst, node)
        return node

    def nameset(self, name: str) -> None:
        self.ast._set(name, self.last_node)

    def nameadd(self, name: str) -> None:
        self.ast._setlist(name, self.last_node)

    def define(
        self,
        keys: Iterable[str],
        list_keys: Iterable[str] | None = None,
    ) -> Any:
        ast = AST()
        ast._define(keys, list_keys=list_keys)
        ast.update(self.ast)
        self.ast = ast
        return self.ast

    @overload
    def __call__(self, node: ParseState) -> Self: ...

    @overload
    def __call__(self, node: Any) -> Any: ...

    @overload
    def __call__(self, **kwargs: Any) -> Any: ...

    def __call__(self, *_args, node: Any = None, **kwargs: Any) -> Any:
        if node is not None and kwargs:
            raise TypeError("Cannot provide both a positional and keyword arguments.")
        if node is None and not kwargs:
            raise TypeError(
                "Must provide either one positional argument or keyword arguments.",
            )

        if isinstance(node, ParseState):
            return self.merge(node)
        elif kwargs:
            for name, value in kwargs.items():
                self.append(value)
                self.nameset(name)
            return self.last_node
        elif node is None:
            return None
        else:
            return self.append(node)


class ParseStateStack:
    __slots__ = ('callstack', 'state_stack')

    def __init__(self, cursor: Cursor) -> None:
        self.state_stack: list[ParseState] = [ParseState(cursor)]
        self.callstack: list[RuleInfo] = []

    def _clone(self) -> Self:
        new = type(self)(self.state.cursor)
        new.state_stack = self.state_stack[:]
        new.callstack = self.callstack[:]
        return new

    @property
    def state(self) -> ParseState:
        return self.state_stack[-1]

    def fold(self) -> Any:  # this is Parsed
        return self.state.fold()

    def undo(self) -> ParseState:
        return self.state_stack.pop()

    def pop(self) -> ParseState:
        prev = self.state_stack.pop()
        self.state.cursor.goto(prev.cursor.pos)
        return prev

    def new(self) -> ParseState:
        newstate = ParseState(self.state.cursor)
        self.state_stack.append(newstate)
        return self.state

    def push(self) -> ParseState:
        newstate = ParseState(self.state)
        self.state_stack.append(newstate)

        return self.state

    def merge(self) -> ParseState:
        prev = self.pop()
        return self.state.merge(prev)

    def alert(self, level: int = 1, message: str = '') -> Alert:
        self.state.alerts.append(Alert(level=level, message=message))
        return self.state.alerts[-1]

    def __copy__(self) -> ParseStateStack:
        new = self.__class__.__new__(self.__class__)

        new.state_stack = self.state_stack[:]
        new.callstack = self.callstack[:]

        return new



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations


type ffset = set[tuple[str, ...]]


def ref(name: str) -> tuple[str]:
    return (_ref(name),)


def kdot(x: ffset, y: ffset, k: int) -> ffset:
    if not y:
        return {a[:k] for a in x}
    elif not x:
        return {b[:k] for b in y}
    else:
        return {(a + b)[:k] for a in x for b in y}


class _ref(str):
    __slots__ = ()

    def __repr__(self) -> str:
        return repr(f'→{self}')



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import dataclasses as dc
from pathlib import Path
from typing import Any

from tatsu import peg as g
from tatsu.util.indent import IndentPrintMixin


def rstype(type: Any, _modelnames: set[str]) -> str:
    type = str(type)
    type = str(type).split('|')[0].strip()
    if type.startswith('tuple['):
        type = type.replace('tuple[', 'Vec<').replace('...]', '>')
    elif type.startswith('dict['):
        type = type.replace('dict[', 'HashMap<').replace(']', '>')
    elif type.startswith('list['):
        type = type.replace('list[', 'Vec<').replace(']', '>')
    else:
        type = type.replace('[', '{').replace(']', '}')
    if type == 'Model':
        type = 'Box<Model>'
    if type == 'Vec<Model>':
        type = 'Vec<Box<Model>>'
    # if type in modelnames:
    #     type = 'Box<dyn Model>'
    # type= type.replace('str', "'static &str")
    type = type.replace('str', "String")
    return type


def main():
    root = Path('./rustsrc')
    root.mkdir(exist_ok=True)

    modelnames = set()
    for cls in g.Model.model_classes():
        name = cls.__name__
        modelnames.add(name)

    n = 0
    for cls in g.Model.model_classes():
        name = cls.__name__
        if (
            name.endswith('Box')
            or 'Comment' in name
            or 'First' in name
            or 'Choice' in name
            or 'Sequence' in name
            or name in {'Option', 'Group'}
        ):
            continue

        fields = []
        for field in dc.fields(cls):  # pyright: ignore[reportArgumentType]
            if not field.init:
                continue
            if field.name.startswith('_'):
                continue
            if field.name in {'ast', 'ctx', 'parseinfo'}:
                continue
            fields.append(field)

        module = cls.__module__.split('.')[-1]
        path = root / f'{module.lower()}.rs'

        n += 1
        p = IndentPrintMixin()
        if not path.is_file():
            p.print("""\
                // Copyright (c) 2026 Juancarlo Añez (apalala@gmail.com)
                // SPDX-License-Identifier: AGPL-3.0-or-later

                use crate::input::Cursor;
                use super::model::{CanParse, ParseResult};
                use crate::engine::{Cst, Ctx};
            """)
        p.print()
        p.print(f'// #{n}')
        p.print(f"""\
            pub struct {name}<M> {{
                pub exp: Box<M>,
            }}
        """)
        p.print()

        # for field in fields:
        #     type = rstype(field.type, modelnames)
        #     with p.indent():
        #         p.print(f'pub {field.name}: {type},')
        #
        # p.print("}")

        p.print(f"""\
            impl<M, C> CanParse<C> for {name}<M>
            where
                M: CanParse<C>,
                C: Cursor,
            {{
                fn parse(&self, mut ctx: Ctx<C>) -> ParseResult<C> {{
                    unimplemented!()
                }}
            }}
        """)
        p.print()
        with path.open('a') as f:
            f.write(p.printed_text())


if __name__ == '__main__':
    main()



#  Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
#  SPDX-License-Identifier: BSD-4-Clause
#

# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from copy import copy
from dataclasses import field
from functools import cached_property
from typing import Any

from ..contexts import Ctx
from ..exceptions import FailedUnlinkedRule
from ..objectmodel import nodedataclass
from ..util import typename
from .base import Grammar, Leaf, Model, Rule
from .syntax import Sequence


@nodedataclass
class RuleInclude(Leaf):
    name: str = field(default='')  # type: ignore
    _exp: Model | None = None

    def __post_init__(self):
        if not self.name and self.ast:
            assert isinstance(self.ast, str)
            self.name = self.ast
        super().__post_init__()
        assert self.name, f'{self!s}'

    def _parse(self, ctx: Ctx) -> Any:
        if self._exp is None:
            raise ctx.newexcept(f'Unlinked rule {self.name!r}', FailedUnlinkedRule)
        return self._exp._parse(ctx)

    @property
    def exp(self) -> Model | None:
        return self._exp

    def link(self, grammar: Grammar) -> None:
        super().link(grammar)
        name = self.name
        assert name and isinstance(name, str), f'{self!r} {self.name!r}'
        self._exp = grammar.rulemap[name].exp
        assert isinstance(self._exp, Model), f'{self!r}\n{self.name!r}\n{self._exp!r}'

    def missing_rules(self, rulenames: set[str]) -> set[str]:
        assert self.name, f'{self!r} {self.name!r}'
        if self.name not in rulenames:
            return {self.name}
        return set()

    def _pretty(self, lean=False):
        _ = lean
        return f'>{self.name}'

    @cached_property
    def defines_single(self) -> list[str]:
        if not self.exp:
            return []
        return self.exp.defines_single

    @cached_property
    def defines_list(self) -> list[str]:
        if not self.exp:
            return []
        return self.exp.defines_list

    def optimized(self) -> Model:
        if not self.exp:
            return super().optimized()
        new = copy(self)  # noqa: F821
        new._exp = self.exp.optimized()
        return new


@nodedataclass
class BasedRule(Rule):
    baserule: Rule = field(default_factory=Rule)
    rhs: Model = Model()

    def __post_init__(self):
        super().__post_init__()

        assert isinstance(
            self.baserule,
            Rule,
        ), f'{typename(self.base)}: {self.baserule=!r}'

        self.params = self.params or self.baserule.params
        self.kwparams = self.kwparams or self.baserule.kwparams

        self.rhs = Sequence(ast=[self.baserule.exp, self.exp])

    def _parse(self, ctx: Ctx) -> Any:
        return self._parse_rhs(ctx, self.rhs)

    @cached_property
    def defines_single(self) -> list[str]:
        return list(set(super().defines_single) | set(self.exp.defines_single))

    @cached_property
    def defines_list(self) -> list[str]:
        return list(set(super().defines_list) | set(self.rhs.defines_list))



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import json
import subprocess  # noqa: S404
import sys

import pytest

from .fixtures import PATH_TATSU_GRAMMAR


pytestmark = pytest.mark.skipif(
    sys.platform == "win32", reason="Does not work on Windows"
)


def uv_run(cmd: list[str]) -> str:
    return subprocess.check_output(
        ['uv', 'run', *cmd],
    ).decode()


CLING = "cling"


def test_cling_help():
    output = uv_run([CLING, '--help'])
    assert '竜TatSu takes a grammar' in output


def test_cling_boot():
    output = uv_run([CLING, 'boot'])
    data = json.loads(output)
    assert data['__class__'] == 'Grammar'
    assert data['name'] == 'TatSuBootstrap'


def test_cling_grammar():
    output = uv_run([CLING, 'grammar', PATH_TATSU_GRAMMAR])
    data = json.loads(output)
    assert data['__class__'] == 'Grammar'
    assert data['name'] == 'TatSu'


def test_cling_grammar_json():
    output = uv_run([CLING, 'grammar', '--json', PATH_TATSU_GRAMMAR])
    data = json.loads(output)
    assert data['__class__'] == 'Grammar'
    assert 'rules' in data



# Copyright © 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .global_opt import add_global_options


def add_run_cmd(subparsers):
    run_parser = subparsers.add_parser(
        "run",
        help="Parse input files with the given grammar",
    )
    add_global_options(run_parser)
    run_parser.add_argument(
        "grammar",
        help="Path to a grammar in EBNF or JSON format",
    )
    run_parser.add_argument(
        "inputs",
        nargs="+",
        help="The files to be parsed",
    )

    format = run_parser.add_mutually_exclusive_group()
    format.add_argument(
        "-j",
        "--json",
        action="store_true",
        dest="json",
        default=True,
        help="Output the grammar in JSON format",
    )
    format.add_argument(
        "-jl",
        "--jsonl",
        action="store_true",
        dest="json_lines",
        default=False,
        help="Output the grammar in JSON Lines format",
    )
    format.add_argument(
        "-M",
        "--model",
        action="store_true",
        dest="model",
        help="Output a Python model of the output",
    )
    run_parser.add_argument(
        "-s", "--start", default="", dest="start", help="Name of the start rule"
    )
    run_parser.add_argument(
        "-n",
        "--nproc",
        type=int,
        default=0,
        dest="nproc",
        help="Number of concurrent workers",
    )
    run_parser.add_argument(
        "-u",
        "--summary",
        action="store_true",
        help="Always show summary (overrides --quiet)",
    )
    return run_parser



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re
import subprocess  # noqa: S404
import sys

import pytest

from .fixtures import PATH_TATSU_GRAMMAR


OTATSU = 'otatsu'

pytestmark = pytest.mark.skipif(
    sys.platform == "win32", reason="Does not work on Windows"
)


def uv_run(cmd: list[str]) -> str:
    return subprocess.check_output(['uv', 'run', *cmd]).decode()


def test_feature_one():
    assert True


def test_cli_help():
    output = uv_run([OTATSU, '--help'])
    pattern = r'(?ms)竜TatSu takes a grammar .*GRAMMAR'
    assert bool(re.search(pattern, output))


def test_cli_python():
    output = uv_run([OTATSU, PATH_TATSU_GRAMMAR])
    pattern = (
        r'(?ms)CAVEAT UTILITOR.*?竜TatSu.*?KEYWORDS = \('
        r'.*?class \w*?Parser\(\w*Parser\):'
    )
    found = re.search(pattern, output)
    assert bool(found), output


def test_cli_model():
    output = uv_run([OTATSU, '-g', PATH_TATSU_GRAMMAR])
    pattern = (
        r'(?ms)CAVEAT UTILITOR.*?竜TatSu'
        r'.*?class \w+?ModelBuilderSemantics\(ModelBuilderSemantics\):'
    )
    assert bool(re.search(pattern, output))



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Any, Protocol

from ..config import ParserConfig
from ..exceptions import FailedLeftRecursion
from ..util.common import is_posix
from ..util.debugging import INFO_print
from ..ztyle import Color, Style
from .ctx import Ctx


if not is_posix():
    C_ENTRY = '<'
    C_SUCCESS = '>'
    C_FAILURE = '!'
    C_RECURSION = 'r '
    C_CUT = '~'
else:
    C_ENTRY = '↙'
    C_SUCCESS = '≡'
    C_FAILURE = '≢'
    C_RECURSION = '⟲\u2005'
    C_CUT = '⚔'


class EventColor:
    def __init__(self, color: Color | None = None):
        if color is None:
            color = Color.tty()
        self.entry = Style(C_ENTRY, bold=True, fg=3, color=color)
        self.success = Style(C_SUCCESS, bold=True, fg=2, color=color)
        self.failure = Style(C_FAILURE, bold=True, fg=1, color=color)
        self.recursion = Style(C_RECURSION, bold=True, fg=1, color=color)
        self.cut = Style(C_CUT, bold=True, fg=5, color=color)
        self.dim = Style(dim=True, color=color)


class Tracer(Protocol):
    def trace(self, msg: str, *args: Any, **kwargs: Any) -> None: ...

    def trace_event(self, ctx: Ctx, event: Style | str) -> None: ...

    def trace_entry(self, ctx: Ctx) -> None: ...

    def trace_success(self, ctx: Ctx) -> None: ...

    def trace_failure(self, ctx, ex: Exception | None = None) -> None: ...

    def trace_recursion(self, ctx: Ctx) -> None: ...

    def trace_cut(self, ctx: Ctx) -> None: ...

    def trace_match(
        self,
        ctx: Ctx,
        token: Any,
        name: str | None = None,
        failed: bool = False,
    ) -> None: ...

    def trace_no_match(
        self,
        ctx: Ctx,
        token: Any,
        name: str | None = None,
    ) -> None:
        self.trace_match(ctx, token, name=name, failed=True)

    def rulestack(self, ctx: Ctx) -> str: ...


class ConsoleTracer(Tracer):
    def __init__(
        self,
        *,
        config: ParserConfig | None = None,
        **settings: Any,
    ) -> None:
        self.config = ParserConfig.new(config, **settings)
        color = Color.tty() if self.config.colorize else Color.never()
        self.ec = EventColor(color)

    def trace(self, msg: str, *args: Any, **kwargs: Any) -> None:
        if not self.config.trace:
            return

        INFO_print(msg, *args, **kwargs)

    def trace_event(self, ctx: Ctx, event: Style | str) -> None:
        if not self.config.trace:
            return

        cursor = ctx.cursor

        source = ''
        if self.config.trace_filename:
            source = cursor.lineinfo().source
        if source:
            source += '\n'

        lookahead = cursor.lookahead().rstrip()

        pos = cursor.lookahead_pos()

        message = (
            f'{event}{self.rulestack(ctx)}'
            f' {self.ec.dim.apply(source)}'
            f'\n{pos}⇥{lookahead}'
        )
        self.trace(message)

    def trace_entry(self, ctx: Ctx) -> None:
        self.trace_event(ctx, self.ec.entry)

    def trace_success(self, ctx: Ctx) -> None:
        self.trace_event(ctx, self.ec.success)

    def trace_failure(self, ctx: Ctx, ex: Exception | None = None) -> None:
        if isinstance(ex, FailedLeftRecursion):
            self.trace_recursion(ctx)
        else:
            self.trace_event(ctx, self.ec.failure)

    def trace_recursion(self, ctx: Ctx) -> None:
        self.trace_event(ctx, self.ec.recursion)

    def trace_cut(self, ctx: Ctx) -> None:
        self.trace_event(ctx, self.ec.cut)

    def trace_match(
        self,
        ctx: Ctx,
        token: Any,
        name: str | None = None,
        failed: bool = False,
    ) -> None:
        if not self.config.trace:
            return

        cursor = ctx.cursor

        name_str = f'/{name}/' if name else ''
        if self.config.trace_filename:
            source = cursor.lineinfo().source + '\n'
        else:
            source = ''

        mark = self.ec.failure if failed else self.ec.success

        lookahead = cursor.lookahead().rstrip()
        lookahead = '\n' + lookahead if lookahead else lookahead

        message = f'{mark}\'{token}{name_str}{self.ec.dim.apply(source)}{lookahead}'
        self.trace(message)

    def rulestack(self, ctx: Ctx) -> str:
        ruleinfos = ctx.callstack
        rulestack = [f'{r.name}' for r in reversed(ruleinfos)]
        stack = self.config.trace_separator.join(rulestack)
        if (
            max((len(s) for s in stack.splitlines()), default=0)
            > self.config.trace_length
        ):
            stack = stack[: self.config.trace_length]
            stack = stack.rsplit(self.config.trace_separator, 1)[0]
            stack += self.config.trace_separator
        return stack


class NullTracer(Tracer):
    def trace(self, msg: str, *args: Any, **kwargs: Any) -> None:
        pass

    def trace_event(self, ctx: Ctx, event: Style | str) -> None:
        pass

    def trace_entry(self, ctx: Ctx) -> None:
        pass

    def trace_success(self, ctx: Ctx) -> None:
        pass

    def trace_failure(self, ctx: Ctx, ex: Exception | None = None) -> None:
        pass

    def trace_recursion(self, ctx: Ctx) -> None:
        pass

    def trace_cut(self, ctx: Ctx) -> None:
        pass

    def trace_match(
        self,
        ctx: Ctx,
        token: Any,
        name: str | None = None,
        failed: bool = False,
    ) -> None:
        pass

    def rulestack(self, ctx: Ctx) -> str:
        _ = ctx
        return ""



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""
ANSI terminal colour — zero dependencies, optionally self-contained.

Provides a ``Color`` policy object and a ``Style`` composer for building
ANSI-wrapped terminal text.  Supports 256-colour palettes, 24-bit RGB,
148 CSS named colours, dynamic TTY detection, ``NO_COLOR`` / ``FORCE_COLOR``
environment variables, and lazy-loading colour maps.

Quick-start:

.. code-block:: python

    from tatsu.ztyle import Color, Style

    style = Style("hello", bold=True, fg=2, color=Color.tty())
    print(style)  # ANSI-wrapped on TTY, plain text otherwise
"""

from __future__ import annotations

from .colormap import *  # noqa: F403
from .csscolormap import *  # noqa: F403
from .style import *  # noqa: F403
from .style import (
    RGB,
    Color,
    Style,
    css_color,
    fmt,
    named_color,
    rgb,
    visual_len,
)


__all__ = [
    "Color",
    "RGB",
    "Style",
    "css_color",
    "fmt",
    "named_color",
    "rgb",
    "visual_len",
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import codecs
import functools
import hashlib
import io
import re
import unicodedata
from collections import namedtuple
from collections.abc import Iterable
from io import StringIO
from typing import Any

from .common import isreserved


def slicetowidth(s: str, width: int) -> str:
    if len(s) > width:
        return s[: width - 3] + '...'
    return s


def linecount(s: str) -> int:
    """
    Return the number of lines in the string using the "editor view".

    In this convention, a string is treated as a sequence of lines separated
    by line breaks. An empty string is considered to contain a single
    empty line (Line 1), ensuring consistency with how IDEs and text
    editors represent buffers.
    """
    return 1 + sum(1 for _ in re.finditer(r"(?m)\r?\n|\r", s))


def ismultiline(s: str) -> int:
    return linecount(s) > 1


lcnt = namedtuple('lcnt', 'totl blnk cmnt code')


def countlines(s: str, cmtstr: str = r'#') -> lcnt:
    """
    SLOC counter with 'Editor View' semantics.
    Memory-efficient streaming via io.StringIO.
    """
    # by Gemini 2026-03-15

    totl = 0
    blnk = 0
    cmnt = 0
    code = 0

    inner = re.compile(rf"""(?x)
        ^ [ \t]* (?:
              (?P<cmnt> [{cmtstr}] )
            | (?P<code>  \S )
        )
    """)

    with io.StringIO(s) as stream:
        for line in stream:
            totl += 1

            if not line:
                blnk += 1
                continue

            match = inner.match(line)
            if not match:
                blnk += 1
            elif match.group('cmnt'):
                cmnt += 1
            else:
                code += 1

    # Editor View adjustment:
    # If the text ends with a break, there is a final ghost line.
    if not totl and s.endswith(('\n', '\r')):
        totl += 1
        blnk += 1

    result = lcnt(totl, blnk, cmnt, code)
    assert totl == blnk + cmnt + code
    return result


def unicode_display_len(text: str) -> int:
    # by Gemini 2026/02/17 (with many amendments)
    """
    Calculates the display width of a string in a terminal or
    fixed-width font context.
    """
    assert isinstance(text, str), repr(text)

    def uwidth(c: str) -> int:
        status = unicodedata.east_asian_width(c)
        return 1 + int(status in {'W', 'F'})

    return sum(uwidth(s) for s in text)


def hasha(text: Any) -> str:
    """
    Generates an SHA-256 hex digest of the provided object.
    """
    # by Gemini (- 2026-02-08)
    # by [apalala@gmail.com](https://github.com/apalala)

    # hashlib requires bytes, so encode the string to UTF-8
    return hashlib.sha256(str(text).encode('utf-8')).hexdigest()


def eval_escapes(s: str | bytes) -> str | bytes:
    """
    Given a string, evaluate escape sequences starting with backslashes as
    they would be evaluated in Python source code. For a list of these
    sequences, see: https://docs.python.org/3/reference/lexical_analysis.html

    This is not the same as decoding the whole string with the 'unicode-escape'
    codec, because that provides no way to handle non-ASCII characters that are
    literally present in the string.
    """
    # by Rob Speer

    escape_sequence_re: re.Pattern = re.compile(r"""(?ux)
        ( \\U........      # 8-digit Unicode escapes
        | \\u....          # 4-digit Unicode escapes
        | \\x..            # 2-digit Unicode escapes
        | \\[0-7]{1,3}     # Octal character escapes
        | \\N\{[^}]+}      # Unicode characters by name
        | \\[\\'"abfnrtv]  # Single-character escapes
        )""")

    def decode_match(match):
        return codecs.decode(match.group(0), 'unicode-escape')

    return escape_sequence_re.sub(decode_match, s)  # type: ignore


def trim(text: str, /, *, all: bool = False, tabwidth: int = 4):
    """
    Trim text of common, leading whitespace.

    Based on the trim algorithm of PEP 257:
        http://www.python.org/dev/peps/pep-0257/
    """
    if not text:
        return ''
    lines = text.expandtabs(tabwidth).splitlines()
    maxindent = len(text)
    indent = maxindent
    start = 0 if all else 1
    for line in lines[start:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    trimmed = [lines[0].strip()] + [line[indent:].rstrip() for line in lines[1:]]
    i = 0
    while i < len(trimmed) and not trimmed[i]:
        i += 1
    return '\n'.join(trimmed[i:])


def indent(text, indent=1, multiplier=4):
    """Indent the given block of text by indent*4 spaces"""
    if text is None:
        return ''
    text = str(text)
    if indent >= 0:
        sindent = ' ' * multiplier * indent
        text = '\n'.join((sindent + t).rstrip() for t in text.splitlines())
    return text


def mangle(name: str) -> str:
    return safe_name(name)


@functools.lru_cache(maxsize=2 * 1024)
def safe_name(name: str, plug: str = "_") -> str:
    """
    Utility to transform a string into a valid Python identifier.
    Raises ValueError for empty inputs or illegal plugs. Handles
    leading digits, whitespace, and reserved hard/soft keywords.
    """
    # with Gemini - January 24, 2026

    if not name:
        raise ValueError("Input string cannot be empty.")
    if not plug or not all(c.isalnum() or c == "_" for c in plug):
        raise ValueError(f"Invalid plug: '{plug}'. Must be non-empty and alphanumeric.")
    if not plug.isidentifier():
        raise ValueError(f"Invalid plug: '{plug}'. Must be valid in identifiers.")

    res: str = name

    if not res.isidentifier():
        res = re.sub(r"\W", plug, name)

    if not res.isidentifier():
        res = ''.join(c if c.isalnum() or c == '_' else plug for c in name)

    if not res.isidentifier():
        if res[0].isdigit():
            res = f"_{res}" if plug[0].isdigit() else f"{plug}{res}"
        else:
            res = f"{plug}{res}"

    assert res.isidentifier(), f"Failed to sanitize '{name}' into '{res}'"

    while isreserved(res):
        res = f"{res}{plug}"

    return res


def pythonize_name(name: str) -> str:
    # by Copilot 2026-03-06
    if not name:
        return name
    # Convert CamelCase to snake_case
    name = re.sub(r'(.)([A-Z][a-z]+)', r'\1_\2', name)
    name = re.sub(r'([a-z0-9])([A-Z])', r'\1_\2', name).lower()
    # Ensure it's a valid Python identifier
    return safe_name(name)


def prints(*args, **kwargs: Any) -> str:
    with StringIO() as f:
        kwargs['file'] = f
        # kwargs['end'] = ''
        print(*args, **kwargs)
        out = f.getvalue()
        return '\n'.join(out.splitlines(False))


def longest_common_prefix(strs: Iterable[str], suffix: str = '') -> str:
    if not strs:
        return ''

    strs = [s + suffix for s in sorted(strs)]
    if len(strs) == 1:
        return strs[0]

    first = strs[0]
    last = strs[-1]

    i = 0
    m = min(len(first), len(last))
    while i < m and first[i] == last[i]:
        i += 1

    return first[:i]


def without_common_prefix(strs: Iterable[str], suffix: str = '') -> list[str]:
    if not strs:
        return []
    prefix = longest_common_prefix(strs, suffix=suffix)
    return [s.lstrip(prefix) for s in strs]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from typing import Protocol, runtime_checkable


@runtime_checkable
class Heart(Protocol):
    def beat(self, mark: int, total: int) -> None: ...
    def dead(self) -> bool:
        return False


class NullHeart(Heart):
    def beat(self, mark: int, total: int) -> None:
        pass



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .._version import version


HEADER: str = f"""\
        #!/usr/bin/env python3
        #
        #  CAVEAT UTILITOR
        #
        #  This file was automatically generated by 竜TatSu v{version}
        #
        #     https://pypi.python.org/pypi/tatsu/
        #
        #  Any changes you make to it will be overwritten the next time
        #  the file is generated.
        #
        # noqa # type: ignore # ruff: noqa
        # copyright: ignore
        # fmt: off

        from __future__ import annotations
    """.rstrip()


IMPORTS: str = """
        from typing import Any

        from tatsu import decorators as tatsu
        from tatsu.contexts import Ctx
        from tatsu.config import ParserConfig
        from tatsu.parsing import Parser
        from tatsu.input.buffer import Buffer
        from tatsu.input.textlines import TextLines
    """.rstrip()


def PARSER_BODY(rules_name: str) -> str:
    return f"""\
        config = ParserConfig.new(config, **settings)
        rulessource = {rules_name}()
        assert isinstance(config, ParserConfig)

        super().__init__(rulessource, config=config)
    """.rstrip()


def FOOTER(name: str) -> str:
    return f"""\
        def main(filename, **kwargs):
            if not filename or filename == '-':
                import sys

                text = sys.stdin.read()
            else:
                import pathlib

                text = pathlib.Path(filename).read_text()

            parser = {name}Parser()
            return parser.parse(text, filename=filename, **kwargs)


        if __name__ == '__main__':
            import json
            from tatsu.parsing import generic_main
            from tatsu.util import asjson

            ast = generic_main(main, {name}Parser, name='{name}')
            data = asjson(ast)
            print(json.dumps(data, indent=2))
    """.rstrip()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import os

import pytest  # noqa


NL = os.linesep
LNL = len(NL)

from tatsu.util.newlines import (
    indent_len,
    take_blankline_len,
    take_dedent_len,
    take_indent_len,
    take_linebreak_len,
)


def test_empty_line():
    assert take_linebreak_len("") == 0
    assert take_linebreak_len("   ") == 3
    assert take_linebreak_len(f"{NL}") == LNL
    assert take_linebreak_len(f"  {NL}next") == 2 + LNL
    assert take_linebreak_len("content") is None


def test_indent_len():
    assert indent_len("no space") == 0
    assert indent_len("  two spaces") == 2
    assert indent_len("\t tab") == 2
    assert indent_len(f"  multi{NL}line") == 2


def test_take_indent():
    # Valid: empty line followed by actual indent
    assert take_indent_len(f"{NL}  code") == 2 + LNL
    # Invalid: empty line followed by zero margin
    assert take_indent_len(f"{NL}no indent") == LNL
    # Invalid: no initial empty line/newline
    assert take_indent_len("code") is None


def test_dedent():
    assert take_dedent_len(f"{NL}code") == LNL
    assert take_dedent_len(f"{NL}  code") is None


def test_blank_line():
    assert take_linebreak_len(f"{NL}") == LNL
    assert take_linebreak_len(f"{NL}{NL}") == LNL
    assert take_linebreak_len(f"{NL}{NL}", 1) == LNL
    assert take_blankline_len(f"{NL}{NL}") == 2 * LNL
    assert take_blankline_len(f"{NL}  {NL}") == 2 + 2 * LNL
    assert take_blankline_len(f"{NL}content") is None


def test_indent_len_with_pos():
    text = "ignore  two spaces"
    # Starting at index 6 ('  two spaces')
    assert indent_len(text, start=6) == 2


def test_indent_with_pos():
    text = f"{NL}  code"
    assert take_indent_len(text) == 2 + LNL
    text = f"ignore{NL}  code"
    assert take_indent_len(text, start=6) == 2 + LNL


def test_blank_line_with_pos():
    # text with a blank line in the middle (start 4 is the second {NL})
    text = f"abc{NL}{NL}def"
    # Starting at the first newline
    assert take_blankline_len(text, start=3) == 2 * LNL
    # Starting at the second newline
    assert take_blankline_len(text, start=4) is None


def test_blank_line_with_whitespace_and_pos():
    # text with a line containing only spaces in the middle
    text = f"abc{NL}    {NL}def"
    # start 4 is the start of the 4 spaces
    assert take_blankline_len(text, start=3) is not None


def test_not_blank_line_with_pos():
    text = f"abc{NL}  code{NL}"
    # start 4 is '  code' - should not be blank
    assert take_blankline_len(text, start=4) is None


def test_blank_line_at_end_with_pos():
    text = f"abc{NL}    "
    # start 4 is the trailing whitespace at the end of the file
    assert take_blankline_len(text, start=4) is not None



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import io
from collections.abc import Generator
from contextlib import contextmanager
from typing import Any

from .abctools import isiter
from .common import typename
from .strtools import ismultiline, trim
from .typetools import notnone


BLACK_LINE_LENGTH = 88


__all__ = ['IndentPrintMixin', 'fold']


class IndentPrintMixin:
    def __init__(self, amount: int = 4, initial: int = 0):
        self.amount = amount
        self.indent_stack: list[int] = [initial]
        self.output_stream = io.StringIO()

    def clear(self):
        self.__init__(amount=self.amount)

    def printed_text(self):
        return self.output_stream.getvalue()

    def print(self, *args, **kwargs):
        args = [trim(str(arg)) for arg in args if arg is not None]
        lines = self.as_printed_lines(*args, **kwargs)
        self._do_print_lines(lines, **kwargs)

    def as_printed(self, *args, **kwargs) -> str:
        lines = self.as_printed_lines(*args, **kwargs)
        return '\n'.join(lines)

    def as_printed_lines(self, *args, **kwargs) -> list[str]:
        text = self.io_print(*args, **kwargs)
        return self.indented_lines(text)

    def indent_amount(self, amount: int = -1, levels: int = 1) -> int:
        assert amount < 0 or amount > 0
        amount = amount if amount > 0 else self.amount
        return amount * levels + self.indentation

    def fitsfmt(self, line: str, addlevels: int = 1):
        assert addlevels >= 0
        if ismultiline(line):
            return False
        total = len(line) + self.indent_amount(levels=addlevels)
        return total <= BLACK_LINE_LENGTH

    def prefixlen(self, addindents: int = 0):
        return addindents * self.amount + self.indentation

    @contextmanager
    def indent(self, amount: int = -1, levels: int = 1) -> Generator:
        self.indent_stack.append(
            self.indent_amount(amount=amount, levels=levels),
        )
        try:
            yield
        finally:
            self.indent_stack.pop()

    @property
    def indentation(self) -> int:
        return self.indent_stack[-1]

    @property
    def indentstr(self) -> str:
        return ' ' * self.indentation

    @property
    def initial(self) -> int:
        return self.indent_stack[0]

    @staticmethod
    def io_print(*args, **kwargs) -> str:
        kwargs.pop('file', None)  # do not allow redirection of output
        with io.StringIO() as output:
            print(*args, file=output, **kwargs)
            return output.getvalue()

    def _do_print_lines(self, lines: list[str] | None = None, **kwargs):
        if not lines:
            print(file=self.output_stream, **kwargs)
            return

        for line in lines:
            print(line, file=self.output_stream, **kwargs)

    def indented_lines(self, text: str) -> list[str]:
        text = trim(text)
        return [
            (self.indentstr + line).rstrip() for line in text.splitlines(keepends=False)
        ]

    def pfold(
        self,
        prefix: str,
        value: Any,
        *,
        sep: str = ',',
        lbrack: str | None = None,
        rbrack: str | None = None,
        reprs: bool = True,
        amount: int = 2,
        initial: int = 0,
        addlevels: int | None = None,
    ) -> None:
        folded = fold(
            prefix,
            value,
            reprs=reprs,
            sep=sep,
            lbrack=lbrack,
            rbrack=rbrack,
            amount=amount,
            initial=initial,
            addlevels=addlevels or 0,
        )
        self.print(folded)


def fold(
    prefix: str,
    value: Any,
    *,
    sep: str = ',',
    lbrack: str | None = None,
    rbrack: str | None = None,
    reprs: bool = True,
    amount: int = 2,
    initial: int = 0,
    addlevels: int = 0,
) -> str:
    brackets = {
        'dict': ('{', '}'),
        'list': ('[', ']'),
        'tuple': ('(', ')'),
        'None': ('', ''),
    }
    lb, rb = brackets.get(typename(value), ('', ''))
    lbrack = notnone(lbrack, lb)
    rbrack = notnone(rbrack, rb)
    assert lbrack is not None and rbrack is not None

    im = IndentPrintMixin(amount=amount, initial=initial)
    if not isiter(value):
        im.print(f'{prefix}{lbrack}{value!r}{rbrack}')
        return im.printed_text().rstrip()

    if isinstance(value, dict):
        if reprs:
            repr_list = [f'{k!r}: {v!r}' for k, v in value.items()]
        else:
            repr_list = [f'{k}: {v}' for k, v in value.items()]
    elif reprs:
        repr_list = [repr(v) for v in value]
    else:
        repr_list = value

    valuestr = f'{prefix}{lbrack}{f'{sep} '.join(repr_list)}{rbrack}'
    if im.fitsfmt(valuestr, addlevels=addlevels):
        im.print(valuestr)
    else:
        im.print(f'{prefix}{lbrack}')
        with im.indent():
            im.print(f'{sep}\n'.join(repr_list))
        im.print(rbrack)
    return im.printed_text().rstrip()



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .cfg import DEFAULT_PYGMENTS_STYLE


def add_global_options(parser):
    group = parser.add_argument_group("global options")
    group.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="Suppress progress bar and spinner output",
    )
    group.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Provide more detailed information about the parsing process",
    )
    group.add_argument(
        "-t",
        "--trace",
        action="store_true",
        help="Display a detailed trace of the parsing process",
    )
    group.add_argument(
        "-o",
        "--output",
        default="",
        help="Output to a file or directory instead of stdout",
    )
    group.add_argument(
        "-c",
        "--color",
        choices=["auto", "always", "never"],
        default="auto",
        help="Control colorized output (default: auto)",
    )
    group.add_argument(
        "-l",
        "--style",
        dest="style",
        default=DEFAULT_PYGMENTS_STYLE,
        help="Pygments style name for syntax highlighting",
    )
    # group.add_argument(
    #     "--profile",
    #     action="store_true",
    #     help="Enable CPU and memory profiling",
    # )



from typing import get_args, get_origin


def format_annotation(tp) -> str:
    origin = get_origin(tp)
    args = get_args(tp)

    if not origin:
        # It's a base type like int or str, just grab its name
        return getattr(tp, "__name__", str(tp))

    # Handle modern Unions (str | int) vs Legacy Unions (typing.Union)
    if origin is getattr(types, "UnionType", None) or str(origin) == "typing.Union":
        return " | ".join(format_annotation(arg) for arg in args)

    # Handle standard generics like list[...] or dict[...]
    origin_name = getattr(origin, "__name__", str(origin))
    arg_strings = ", ".join(format_annotation(arg) for arg in args)
    return f"{origin_name}[{arg_strings}]"


# Test it out on a deeply nested messy type object
complex_dynamic_type = list[dict[str, int | float]]
print(format_annotation(complex_dynamic_type))
# Output: list[dict[str, int | float]]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Callable
from functools import cached_property
from typing import Protocol, Self, runtime_checkable

from .infos import LineInfo


@runtime_checkable
class Cursor(Protocol):
    pos: int
    textstr: str

    def clone(self) -> Self: ...

    @property
    def input(self) -> Text: ...

    @property
    def source(self) -> str: ...

    @property
    def linecount(self) -> int: ...

    @property
    def line(self) -> int: ...

    @property
    def current(self) -> str | None: ...

    @property
    def token(self):
        return self.current

    @cached_property
    def namechars(self) -> set[str]: ...

    def goto(self, pos) -> None: ...
    def move(self, n: int) -> None: ...
    def atend(self) -> bool: ...
    def ateol(self) -> bool: ...
    def next(self) -> str | None: ...

    def next_token(self) -> None: ...
    def match(self, token: str) -> str | None: ...
    def matchre(self, pattern: str) -> str | None: ...
    def matcheol(self) -> bool: ...
    def matchname(self) -> str | None: ...
    def matchint(self) -> int | None: ...
    def matchuint(self) -> int | None: ...
    def matchfloat(self) -> float | None: ...
    def matchbool(self) -> bool | None: ...

    def is_name(self, s: str) -> bool: ...
    def is_name_char(self, c: str | None) -> bool: ...

    def lineat(self, pos: int | None = None) -> int: ...
    def lineinfo(self, pos: int | None = None) -> LineInfo: ...

    def lookahead(self) -> str: ...
    def lookahead_pos(self) -> str: ...


@runtime_checkable
class Text(Protocol):
    def newcursor(self) -> Cursor: ...

    @property
    def source(self) -> str: ...


Tokenizer = Text


def match_name(s: str, pos: int, namechars: set[str]) -> int:
    def is_name_char(c: str) -> bool:
        return bool(c) and (c == '_' or c.isalnum() or c in namechars)

    if not s or pos < 0 or pos >= len(s):
        return -1

    p = pos
    is_name_start = (c := s[p]) and (c == '_' or c.isalpha() or c in namechars)
    if not is_name_start:
        return -1
    p += 1

    while p < len(s) and is_name_char(s[p]):
        p += 1

    return p


def matchname(c: Cursor) -> str | None:
    if (p := match_name(c.textstr, c.pos, c.namechars)) <= 0:
        return None
    i = c.pos
    c.goto(p)
    out = c.textstr[i:p]
    return out


def match_bool(s: str, pos: int) -> int:
    if not s or pos < 0 or pos >= len(s):
        return -1
    rest = s[pos:]
    if rest.startswith('true') or rest.startswith('True'):
        return pos + len('true')
    if rest.startswith('false') or rest.startswith('False'):
        return pos + len('false')
    return -1


def matchbool(c: Cursor) -> bool | None:
    if (p := match_bool(c.textstr, c.pos)) is None:
        return None
    i = c.pos
    c.goto(p)
    return bool(c.textstr[i:p].capitalize())


def match_uint(s: str, pos: int) -> int:
    """Matches an integer with optional sign and internal underscores."""
    p = pos
    while p < len(s):
        c = s[p]
        if c.isdigit():
            p += 1
        elif c == '_':
            if p + 1 < len(s) and s[p + 1].isdigit():
                p += 1
            else:
                return -1
        elif c.isalpha():
            return -1
        else:
            break
    return p


def match_int(s: str, pos: int) -> int:
    p = pos
    if p < len(s) and s[p] in {'+', '-'}:
        p += 1

    if p >= len(s) or not s[p].isdigit():
        return -1

    return match_uint(s, p)


def match_float(s: str, pos: int) -> int:
    if not s or pos < 0 or pos >= len(s):
        return -1

    p = pos
    if (p := match_int(s, p)) <= 0:
        return -1

    if p < len(s) and s[p] == '.':
        p += 1
        if (q := match_int(s, p)) > 0:
            p = q

    if p < len(s) and s[p].lower() == 'e':
        p += 1
        if (p := match_int(s, p)) <= 0:
            return -1
    return p


def matchstr(c: Cursor, match: Callable[[str, int], int]) -> str | None:
    if (p := match(c.textstr, c.pos)) <= 0:
        return None
    i = c.pos
    c.goto(p)
    return c.textstr[i:p]


def matchint(c: Cursor) -> int | None:
    if (s := matchstr(c, match_int)) is not None:
        return int(s)
    return None


def matchuint(c: Cursor) -> int | None:
    if (s := matchstr(c, match_uint)) is not None:
        return int(s)
    return None


def matchsigned(c: Cursor) -> int | None:
    if (s := matchstr(c, match_int)) is not None:
        return int(s)
    return None


def matchfloat(c: Cursor) -> float | None:
    if (s := matchstr(c, match_float)) is not None:
        return float(s)
    return None



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import time
from collections.abc import Callable, Generator, Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from ..util import countlines, debugging, slicetowidth
from ..ztyle import Color
from .payload import StrPayload
from .result import Result


type PrintFunc = Callable[..., None]


@dataclass
class ParseStats:
    file_count: int = 0
    totl_lines: int = 0
    code_lines: int = 0
    cmnt_lines: int = 0
    blnk_lines: int = 0
    succ_count: int = 0
    succ_lines: int = 0
    succ_rate: float = 0.0
    run_time: float = 0.0
    slocs_avg: float = 0.0


class SummaryStyle:
    def __init__(self, color: Color):
        s = color.style()
        self.label = s.cyan().dim()
        self.plain = s
        self.good = s.green()
        self.bad = s.red()
        self.warn = s.yellow()
        self.bold_good = s.green().bold()
        self.bold_bad = s.red().bold()


class ResultsStyle:
    def __init__(self, color: Color | None = None):
        color = color or Color.default()
        s = color.style()
        self.header = s.cyan().dim()
        self.good = s.green()
        self.bad = s.red()
        self.plain = s.white().dim()


def format_duration(seconds: float, fractions: bool = False) -> str:
    h = int(seconds // 3600)
    m = int((seconds % 3600) // 60)
    s = seconds % 60
    if h >= 1:
        if fractions:
            return f"{h}:{m:02d}:{s:05.2f}"
        return f"{h}:{m:02d}:{s:02.0f}"
    if fractions:
        return f"{m}:{s:05.2f}"
    return f"{m}:{s:02.0f}"


def result_stats(stats: ParseStats, results: Iterable[Result]) -> Iterable[Result]:
    eolcmt = {
        ".java": "//",
        ".py": "#",
        ".rs": "//",
        ".go": "//",
        ".js": "//",
        ".ts": "//",
    }
    for r in results:
        stats.file_count += 1
        stats.run_time += r.runtime

        p = r.payload
        if not hasattr(r.payload, 'path'):
            p = StrPayload(r.payload)

        suffix = p.path.suffix
        counts = countlines(p.payload, eolcmt.get(suffix, "//"))
        stats.totl_lines += counts.totl
        stats.code_lines += counts.code
        stats.cmnt_lines += counts.cmnt
        stats.blnk_lines += counts.blnk

        if r.success and not r.exception:
            stats.succ_count += 1
            stats.succ_lines += counts.totl
        yield r

    stats.slocs_avg = stats.totl_lines / stats.run_time if stats.run_time > 0 else 0
    stats.succ_rate = stats.succ_count / stats.file_count if stats.file_count else 0


def show_summary(
    start_time: float,
    results: Iterable[Result],
    /,
    eprint: PrintFunc = debugging.eprint,
    *,
    usecolor: bool = True,
    verbose: bool = True,
) -> Generator[Result, None, None]:
    s = SummaryStyle(Color(usecolor))

    st = ParseStats()
    results = result_stats(st, results)
    results = list(results)
    fail = st.file_count - st.succ_count

    results = list(results)
    out: list[Result] = []
    for r in results:
        out.append(r)
        if verbose and r.exception:
            eprint()
            eprint(r.exception)

    if verbose:
        mk = s.bold_bad if fail else s.bold_good
        eprint()
        eprint(mk(f"FAILURES: {fail}" if fail else f"NO FAILURES: {fail}"))

    W = 18

    def row(la: str, va: Any) -> str:
        return f"  {s.label(f'{la:>{W}}')}  {va}"

    eprint()
    eprint(row("files input", s.plain(f"{st.file_count:>12}")))
    eprint(row("source lines input", s.plain(f"{st.totl_lines:>12}")))
    eprint(row("success lines", s.plain(f"{st.succ_lines:>12}")))
    eprint(row("sloc", s.plain(f"{st.code_lines:>12}")))

    rate_st = s.good if st.succ_rate >= 1.0 else s.warn if st.succ_rate > 0.6 else s.bad
    eprint(row("successes", s.good(f"{st.succ_count:>12}")))
    eprint(row("failures", s.bad(f"{fail:>12}")))
    eprint(row("success rate", rate_st(f"{100.0 * st.succ_rate:>12.0f} %")))

    sloc_st = (
        s.good if st.slocs_avg >= 200 else s.warn if st.slocs_avg >= 180 else s.bad
    )
    eprint(row("sloc/sec", sloc_st(f"{st.slocs_avg:>12,.0f} sl/s")))

    eprint(row("run time", s.plain(format_duration(st.run_time, False).rjust(12))))
    eprint(
        row(
            "wall time",
            s.plain(format_duration(time.time() - start_time, False).rjust(12)),
        )
    )
    eprint()
    yield from out


def show_result(
    rprint: PrintFunc, r: Result, /, *, width: int = 60, usecolor: bool = False
) -> None:
    c = Color(usecolor)
    s = ResultsStyle(c)
    nm = s.plain(slicetowidth(Path(r.payload.path).name, width))
    if r.exception or isinstance(r.outcome, Exception):
        rprint(f" {s.bad('✗')} {nm:{width}}{s.bad(f'⏲ {r.runtime:>7.3f}')}")
    else:
        rprint(f" {s.good('✓')} {nm:{width}}{s.good(f'⏲ {r.runtime:>7.3f}')}")


def show_results(
    rprint: PrintFunc,
    results: Iterable[Result],
    /,
    *,
    usecolor: bool = False,
) -> Iterable[Result]:
    for r in results:
        show_result(rprint, r, usecolor=usecolor)
        yield r



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import argparse
import sys
from importlib import resources
from pathlib import Path

from .. import peg as g
from ..api import compile
from ..util import cast
from .semantics import ANTLRSemantics


__all__ = ['translate', 'add_argparse_options', 'main']


def antlr_grammar() -> str:
    path = resources.files('tatsu.g2e') / 'antlr.tatsu'
    return path.read_text()


def translate(
    text: str | None = None,
    filename: str | None = None,
    name: str | None = None,
    encoding: str = 'utf-8',
    trace: bool = False,
    recursion_limit: int | None = None,
) -> g.Grammar:
    if text is None and filename is None:
        raise ValueError('either `text` or `filename` must be provided')
    filepath = Path(filename) if filename else Path()

    if text is None:
        name = name or filepath.stem
        text = filepath.read_text(encoding=encoding)

    name = name or 'Unknown'

    if recursion_limit is not None:
        sys.setrecursionlimit(recursion_limit)

    semantics = ANTLRSemantics(name)
    grammar = compile(antlr_grammar())
    model = grammar.parse(
        text,
        name=name,
        filename=filename,
        semantics=semantics,
        trace=trace,
        colorize=True,
    )
    return cast(g.Grammar, model)


def add_g2e_cmd(subparsers) -> argparse.ArgumentParser:
    g2e_parser = subparsers.add_parser(
        "g2e",
        help="Translate ANTLR grammars to TatSu",
    )
    add_argparse_options(g2e_parser)
    return g2e_parser


def add_argparse_options(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "filename",
        type=Path,
        help="ANTLR grammar file to translate",
    )
    parser.add_argument(
        "-t",
        "--trace",
        action="store_true",
        help="Produce verbose parsing output",
    )
    parser.add_argument(
        "--recursion-limit",
        type=int,
        default=None,
        help="Set the Python recursion limit",
    )


def g2e_cmd(parser: argparse.ArgumentParser) -> int:
    args = parser.parse_args()

    model = translate(
        filename=str(args.filename),
        trace=args.trace,
        recursion_limit=args.recursion_limit,
    )
    print(model.pretty())
    return 0


def main() -> int:
    parser = argparse.ArgumentParser(
        description="Translate ANTLR grammars to TatSu",
    )
    add_argparse_options(parser)
    return g2e_cmd(parser)


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from collections.abc import Callable
from typing import Any, NamedTuple, Protocol

from ..input import Cursor


# NOTE duplicate definition to avoid circular dependencies
type Func = Callable[[Any], Any]


class MemoKey(NamedTuple):
    pos: int
    ruleinfo: RuleInfo


class RuleResult(NamedTuple):
    node: Any
    newpos: int


class RuleLike(Protocol):
    no_memo: bool = False
    no_stak: bool = False
    is_name: bool = False
    is_tokn: bool = False
    is_memo: bool = True
    is_lrec: bool = False

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        pass


class RuleInfo(NamedTuple):
    name: str
    instance: Any
    func: Callable[..., Any]
    no_memo: bool
    no_stak: bool
    is_name: bool
    is_tokn: bool

    is_lrec: bool
    is_memo: bool

    params: tuple[Any, ...]
    kwparams: dict[str, Any]

    @staticmethod
    def new(instance: Any, func: Callable, params=None, kwparams=None) -> RuleInfo:
        name = getattr(func, '__name__', '<?>')
        is_tokn = name.lstrip('_')[:1].isupper()
        return RuleInfo(
            name=name,
            instance=instance,
            func=func,
            no_memo=getattr(func, 'no_memo', False),
            no_stak=getattr(func, 'no_stak', False),
            is_name=getattr(func, 'is_name', False),
            is_tokn=getattr(func, 'is_tokn', is_tokn),
            is_lrec=getattr(func, 'is_lrec', False),
            is_memo=getattr(func, 'is_memo', True),
            params=params or (),
            kwparams=kwparams or {},
        )

    @staticmethod
    def bind(ri: RuleInfo, instance: Any, func: Func | None = None) -> RuleInfo:
        if func is not None:
            return ri._replace(instance=instance, func=func)
        else:
            return ri._replace(instance=instance)

    @property
    def memoizable(self) -> bool:
        return self.is_memo and not self.no_memo

    @property
    def should_trace(self) -> bool:
        return not self.no_stak

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, RuleInfo):
            return self.name == other.name
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


class ParseInfo(NamedTuple):
    cursor: Cursor
    rule: str
    pos: int
    endpos: int
    line: int
    endline: int
    alerts: list[Alert] = []  # noqa: RUF012


class CommentInfo(NamedTuple):
    inline: list
    eol: list

    @staticmethod
    def new_comment():
        return CommentInfo([], [])


class Alert(NamedTuple):
    level: int = 1
    message: str = ''



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .basenode import (
    BaseNode,
    NodeDataclassParams,
    NodeDataclassParams as TatSuDataclassParams,
    nodedataclass,
    nodedataclass as tatsudataclass,
)
from .builder import (
    BuilderConfig,
    ModelBuilderSemantics,
    TypeResolutionError,
)
from .node import Node
from .synth import SynthNode, registered_synthetics, synthesize


__all__ = [
    'BaseNode',
    'Node',
    'SynthNode',
    'NodeDataclassParams',
    'nodedataclass',
    'registered_synthetics',
    'synthesize',
    'tatsudataclass',
    'TatSuDataclassParams',
    'BuilderConfig',
    'ModelBuilderSemantics',
    'TypeResolutionError',
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
"""Progress bar row with lifecycle (NEW→RUNNING→STOPPED) and rendering."""

from __future__ import annotations

from enum import IntEnum, auto
from typing import Any

from ..packetz import WithID
from ..util import clock_time_μs  # noqa: PLC2403
from ..ztyle import Style
from .col import Col
from .metrics import Metrics


__all__ = ["BarRow", "State"]


class State(IntEnum):
    NEW = auto()
    RUNNING = auto()
    STOPPING = auto()
    STOPPED = auto()


class BarRow(WithID):
    """A lightweight, fully picklable data object given to the user."""

    def __init__(
        self,
        *,
        pos: int = 0,
        total: int = -1,
        label: str | Style = "",
        width: int = 0,
        fill: str = "=>.",
        style: list[Style] | None = None,
        cols: list[Any] | None = None,
        selfstop: bool = True,
    ):
        self.begun: int = 0
        self.state: State = State.NEW
        self.selfstop: bool = selfstop

        self.pos: int = pos
        self.total: int = max(1, total)

        self.width: int = width
        self.fill = fill
        self.style: list[Style] = style or []

        self.label = label
        self._cols: list[Any] = []
        if cols is not None:
            self._cols = cols
        elif label:
            self._cols = [Col.label, Col.bar]
        else:
            self._cols = [Col.bar]

    @property
    def cols(self) -> list[Any]:
        return self._cols

    @property
    def key(self) -> float:
        if self.total <= 0:
            return 1
        else:
            return self.total / (1 + self.pos)

    def start(self) -> None:
        self.state = State.RUNNING
        self.begun = clock_time_μs()

    def stop(self) -> None:
        if self.state == State.NEW:
            self.state = State.STOPPING
        else:
            self.state = State.STOPPED

    def is_new(self) -> bool:
        return self.state == State.NEW

    def is_active(self) -> bool:
        return (
            self.has_started() and not self.is_stopped()
        ) or 1 < self.pos < self.total

    def has_started(self) -> bool:
        return self.state == State.RUNNING

    def is_stopped(self) -> bool:
        return self.state in {State.STOPPED, State.STOPPING} or self.pos >= self.total

    def is_stopping(self) -> bool:
        return self.state == State.STOPPING

    def snap(self) -> dict[str, Any]:
        return {
            name: value
            for name, value in vars(self).items()
            if not name.startswith("_")
        }

    def update(
        self,
        /,
        pos: int,
        total: int = -1,
        label: str | None = None,
        fill: str | None = None,
        style: list[Style] | None = None,  # noqa: ARG002
        cols: list[Any] | None = None,  # noqa: ARG002
        *_args,
        **_kwargs,
    ):
        """Write-only operation from the user's side."""
        if total > 0:
            self.total = total
        self.pos = max(0, min(pos, self.total))
        if label:
            self.label = label
        if fill is not None:
            self.fill = fill

    def render(self, m: Metrics) -> list[Any]:
        return [m.resolve(c) if isinstance(c, Col) else c for c in self._cols]

    def metrics(self) -> Metrics:
        return Metrics(self)

    def _call_render(self) -> list[Any]:
        if self.is_stopping() or (self.selfstop and self.pos >= self.total):
            self.stop()

        if self.is_stopped():
            return []

        return self.render(self.metrics())



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from functools import cached_property
from typing import Any

from ..contexts import Ctx
from ..objectmodel import nodedataclass
from .base import Leaf
from .math import ffset


EOF_SYM = '$'
EOL_SYM = '⏎'


@nodedataclass
class Dot(Leaf):
    def _parse(self, ctx: Ctx) -> Any:
        return ctx.dot()

    def _pretty(self, lean=False):
        _ = lean
        return '/./'

    def _first(self, k: int, f: dict[str, ffset]) -> ffset:
        _ = k
        _ = f
        return {('.',)}


@nodedataclass
class Fail(Leaf):
    def _parse(self, ctx: Ctx) -> Any:
        return ctx.fail()

    def _pretty(self, lean=False):
        _ = lean
        return '!()'


@nodedataclass
class Comment(Leaf):
    comment: str = ''

    def _pretty(self, lean: bool = False):
        _ = lean
        return f'(* {self.comment} *)'


@nodedataclass
class EOLComment(Comment):
    def _pretty(self, lean=False):
        _ = lean
        return f'  # {self.comment}\n'


@nodedataclass
class EOF(Leaf):
    def __post_init__(self):
        super().__post_init__()
        self.ast = None

    def _parse(self, ctx: Ctx) -> Any:
        ctx.eofcheck()

    def _pretty(self, lean=False):
        _ = lean
        return '$'


@nodedataclass
class EOL(Leaf):
    def __post_init__(self):
        super().__post_init__()
        self.ast = None

    def _parse(self, ctx: Ctx) -> Any:
        ctx.eolcheck()

    @cached_property
    def _nullable(self) -> bool:
        return True

    def _first(self, k: int, f: dict[str, ffset]) -> ffset:
        _ = k
        _ = f
        return {('$->',)}

    def _pretty(self, lean=False):
        _ = lean
        return EOL_SYM


@nodedataclass
class Token(Leaf):
    token: str = ''

    def __post_init__(self):
        super().__post_init__()
        self.token = self.token or self.ast

    def _parse(self, ctx: Ctx) -> Any:
        return ctx.token(self.token)

    def _first(self, k: int, f: dict[str, ffset]) -> ffset:
        _ = k
        _ = f
        return {(self.token,)}

    def _pretty(self, lean=False):
        _ = lean
        return repr(self.token)


@nodedataclass
class Constant(Leaf):
    literal: str = ''

    def __post_init__(self):
        super().__post_init__()
        self.literal = self.literal or self.ast

    def _parse(self, ctx: Ctx) -> Any:
        return ctx.constant(self.literal)

    def _first(self, k: int, f: dict[str, ffset]) -> ffset:
        _ = k
        _ = f
        return {()}

    def _pretty(self, lean=False):
        _ = lean
        return f'`{self.literal!s}`'

    @cached_property
    def _nullable(self) -> bool:
        return True


@nodedataclass
class Alert(Constant):
    level: int = 0

    def __post_init__(self):
        super().__post_init__()
        if self.ast:
            self.literal = self.ast.message.literal
            self.level = len(self.ast.level)

    def _parse(self, ctx: Ctx) -> Any:
        return ctx.alert(message=self.literal, level=self.level)

    def _pretty(self, lean=False):
        _ = lean
        return f'{"^" * self.level}{super()._pretty()}'


@nodedataclass
class Cut(Leaf):
    def __post_init__(self):
        super().__post_init__()
        self.ast = None

    def _parse(self, ctx: Ctx) -> Any:
        ctx.cut()

    def _first(self, k: int, f: dict[str, ffset]) -> ffset:
        _ = k
        _ = f
        return {()}

    def _pretty(self, lean=False):
        _ = lean
        return '~'

    @cached_property
    def _nullable(self) -> bool:
        return True



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import pytest

from tatsu import compile
from tatsu.exceptions import FailedParse


def test_basic_eol():
    grammar = r"""
        start = 'hello' $-> 'world' ;
    """
    model = compile(grammar)
    ast = model.parse('hello\nworld')
    assert ast == ['hello', 'world']

    ast = model.parse('hello  \nworld')
    assert ast == ['hello', 'world']

    with pytest.raises(FailedParse):
        model.parse('hello world')  # No line break

    with pytest.raises(FailedParse):
        model.parse('helloX\nworld')  # Non-whitespace before EOL


def test_eol_at_end_of_text():
    grammar = r"""
        start = 'hello' $-> $ ;
    """
    model = compile(grammar)
    ast = model.parse('hello\n')
    assert ast == 'hello'

    ast = model.parse('hello  \n')
    assert ast == 'hello'

    with pytest.raises(FailedParse):
        model.parse('hello world')  # No line break at end


def test_multiple_eols():
    grammar = r"""
        start = 'line1' $-> 'line2' $-> 'line3' ;
    """
    model = compile(grammar)
    ast = model.parse('line1\nline2\nline3')
    assert ast == ['line1', 'line2', 'line3']

    ast = model.parse('line1  \nline2\n  line3')
    assert ast == ['line1', 'line2', 'line3']


def test_eol_with_indentation():
    grammar = r"""
        start = 'start' $-> 'indented' $-> 'end' ;
    """
    model = compile(grammar)
    ast = model.parse('start\n  indented\nend')
    assert ast == ['start', 'indented', 'end']

    model.parse('start\nindented\nend')  # Missing indentation
    assert ast == ['start', 'indented', 'end']


def test_eol_in_closure():
    grammar = r"""
        start = { item $-> }* 'end' ;
        item = 'item' ;
    """
    model = compile(grammar)
    ast = model.parse('item\nitem\nend')
    assert ast == [['item', 'item'], 'end']

    ast = model.parse('item  \nitem\nend')
    assert ast == [['item', 'item'], 'end']

    ast = model.parse('end')
    assert ast == [[], 'end']


def test_eol_with_comments():
    grammar = r"""
        @@eol_comments :: /(?m)#.*$/
        start = 'hello' $-> 'world' ;
    """
    model = compile(grammar)

    ast = model.parse('hello # comment\nworld')
    assert ast == ['hello', 'world']

    ast = model.parse("""
        hello   # another comment

        world
        """)
    assert ast == ['hello', 'world']

    ast = model.parse('hello # comment\n# another comment\nworld')
    assert ast == ['hello', 'world']

    ast = model.parse('hello \n  # another comment\n\nworld')
    assert ast == ['hello', 'world']


def test_eol_with_mixed_whitespace():
    grammar = r"""
        start = 'start' $-> 'next' ;
    """
    model = compile(grammar)
    # Test with spaces and tabs
    ast = model.parse('start \t \nnext')
    assert ast == ['start', 'next']

    # Test with only spaces
    ast = model.parse('start   \nnext')
    assert ast == ['start', 'next']

    # Test with only tabs
    ast = model.parse('start\t\nnext')
    assert ast == ['start', 'next']


def test_eol_no_whitespace_before_linebreak():
    grammar = r"""
        start = 'start' $-> 'next' ;
    """
    model = compile(grammar)
    ast = model.parse('start\nnext')
    assert ast == ['start', 'next']


def test_eol_followed_by_non_whitespace():
    grammar = r"""
        start = 'start' $-> 'next' ;
    """
    model = compile(grammar)
    with pytest.raises(FailedParse):
        model.parse('startX\nnext')


def test_eol_followed_by_non_linebreak():
    grammar = r"""
        start = 'start' $-> 'next' ;
    """
    model = compile(grammar)
    with pytest.raises(FailedParse):
        model.parse('start next')


# This test attempts to reproduce the issue with ENDRULE in tatsu.ebnf
# The actual tatsu.ebnf grammar is complex, so we'll simplify it
# to focus on the $-> behavior within a rule definition context.
# This assumes that the problem is with the $-> expression itself
# and not other parts of the tatsu.ebnf grammar.


# Simplified grammar that mimics the structure of ENDRULE
# ENDRULE in tatsu.ebnf is defined as:
# ENDRULE = (';' | $->) ;
# This means it should match either a semicolon or an EOL.
def test_eol_in_tatsu_ebnf_endrule():
    grammar = r"""
        start = { +=definition ENDRULE }+  $ ;
        definition = name ;
        name = /\w+/ ;
        ENDRULE = $-> $-> | $ | ';' ;
    """
    model = compile(grammar)

    # Test with semicolon
    ast = model.parse('rule_name \n\nsome_definition ; next_rule')
    assert ast == ['rule_name', 'some_definition', 'next_rule']

    # Test with EOL
    ast = model.parse('rule_name ; some_definition \n\nnext_rule')
    assert ast == ['rule_name', 'some_definition', 'next_rule']

    ast = model.parse('rule_name \n\nsome_definition   \n\nnext_rule')
    assert ast == ['rule_name', 'some_definition', 'next_rule']

    # Test failure if neither matches
    with pytest.raises(FailedParse):
        model.parse('rule_name some_definition next_rule')

    # Test with a more complex scenario where $-> is part of a choice
    grammar_complex_endrule = r"""
        start = 'rule_name' definition (';' | $->) 'next_rule' ;
        definition = 'some_definition' ;
    """
    model_complex = compile(grammar_complex_endrule)

    ast = model_complex.parse('rule_name some_definition \nnext_rule')
    assert ast == ['rule_name', 'some_definition', 'next_rule']

    ast = model_complex.parse('rule_name some_definition ; next_rule')
    assert ast == ['rule_name', 'some_definition', ';', 'next_rule']

    # If the problem persists with the actual tatsu.ebnf, it might be due to
    # interaction with other rules or whitespace handling in the main grammar.
    # This test confirms the basic functionality of $-> within a choice.



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from functools import cached_property
from typing import Self

from . import LineInfo
from .cursor import Cursor, Text
from .infos import LineIndexInfo


class NullCursor(Cursor):
    def __init__(self) -> None:
        self.pos = 0
        self.textstr = ''
        self._namechars: set[str] = set()

    def clone(self) -> Self:
        return type(self)()

    @property
    def input(self) -> Text:
        return NullText()

    @property
    def linecount(self) -> int:
        return 1  # note: editor view

    @property
    def source(self) -> str:
        return ''

    @property
    def ignorecase(self) -> bool:
        return False

    @property
    def line(self) -> int:
        return 0

    @cached_property
    def namechars(self) -> set[str]:
        return self._namechars

    def goto(self, pos) -> None:
        _ = pos

    def move(self, n: int) -> None:
        _ = n

    def atend(self) -> bool:
        return False

    def ateol(self) -> bool:
        return False

    @property
    def current(self) -> str | None:
        return None

    @property
    def token(self):
        return self.current

    def next(self) -> str | None:
        return None

    def next_token(self) -> None:
        return None

    def match(self, token: str) -> str | None:
        _ = token
        return None

    def matchre(self, pattern: str) -> str | None:
        _ = pattern
        return None

    def matchname(self) -> str | None:
        return None

    def matchint(self) -> int | None:
        return None

    def matchuint(self) -> int | None:
        return None

    def matchfloat(self) -> float | None:
        return None

    def matcheol(self) -> bool:
        return False

    def matchbool(self) -> bool | None:
        return None

    def is_name(self, s: str) -> bool:
        _ = s
        return False

    def is_name_char(self, c: str | None) -> bool:
        _ = c
        return False

    def lineat(self, pos: int | None = None) -> int:
        _ = pos
        return 0

    def lineinfo(self, pos: int | None = None) -> LineInfo:
        _ = pos
        return LineInfo('', 0, 0, 0, 0, '')

    def get_line(self, n: int | None = None) -> str:
        _ = n
        return ''

    def get_lines(self, start: int | None = None, end: int | None = None) -> list[str]:
        _ = start
        _ = end
        return []

    def line_index(self, start: int = 0, end: int | None = None) -> list[LineIndexInfo]:
        _ = start
        _ = end
        return []

    def lookahead(self) -> str:
        return ''

    def lookahead_pos(self) -> str:
        return ''


class NullText(Text):
    def newcursor(self, pos: int = 0) -> Cursor:
        _ = pos
        return NullCursor()

    @property
    def text(self) -> str:
        return ''

    @property
    def source(self) -> str:
        return ''



class DocumentID(str):
    def __new__(cls, content, owner):
        # __new__ handles the string creation
        return super().__new__(cls, content)

    def __init__(self, content, owner):
        # __init__ handles extra state attributes
        self.owner = owner


# Usage
doc = DocumentID("abc-123", "Alice")
print(doc)  # Output: abc-123 (behaves like a string)
print(doc.owner)  # Output: Alice (custom attribute)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import tatsu
from tatsu.util.strtools import trim


def test_multiline_string():
    grammar = r'''
        @@grammar :: Test

        start: longone | shortone $

        shortone: "short"

        # longone: 'long'
        longone: """
            this "" \" \"""
            is a long "string"
            """
    '''
    m = tatsu.compile(grammar)
    m.parse('short')
    input = trim("""
            this "" \" \"""
            is a long "string"
            """).strip()
    m.parse(input)



import time
from dataclasses import dataclass, field


@dataclass(slots=True, kw_only=True)
class Bar:
    name: str
    total: int = 100
    current: int = 0
    start_time: float = field(default_factory=time.time)

    def advance(self, delta: int = 1):
        self.current = min(self.current + delta, self.total)

    def render(self, width: int, metrics: dict[str, float]) -> list[Style]:
        """
        The Default Protocol.
        Returns a rich list of Style components. The manager expands '{bar}'
        using the leftover column width budget.
        """
        color = Color.default()

        # Build pieces using your rich Style architecture
        return [
            color.style(f"{self.name:<20}", bold=True),
            color.style(" "),
            color.style("{bar}"),  # Structural marker
            color.style(" "),
            color.style(f"{metrics['pct']:>3.0f}%", fg=2),  # Green index text
        ]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import dataclasses as dc
import json
import weakref
from collections.abc import Iterable, Mapping, MutableMapping
from typing import Any, Self

from ..util.fromjson import fromjson
from .basenode import BaseNode, nodedataclass


__all__ = ['Node', 'nodedataclass']


_children_cache: MutableMapping[Node, tuple[Node, ...]] = weakref.WeakKeyDictionary()


@nodedataclass
class Node(BaseNode):
    _parent_ref: weakref.ref[Node] | None = dc.field(init=False, default=None)

    def __init__(self, ast: Any = None, **kwargs: Any):
        super().__init__(ast=ast, **kwargs)
        self._parent_ref = None

    @property
    def parent(self) -> Node | None:
        ref = self._parent_ref
        if ref is not None:
            # noinspection PyCallingNonCallable
            return ref()
        else:
            return None

    @property
    def text(self) -> str | None:
        pi = self.parseinfo
        if pi and hasattr(pi.cursor, "text"):
            return pi.cursor.textstr[pi.pos : pi.endpos]
        return None

    @property
    def line(self) -> int | None:
        return self.parseinfo.line if self.parseinfo else None

    @property
    def path(self) -> tuple[Node, ...]:
        ancestors: list[Node] = [self]
        parent = self.parent
        while parent is not None:
            ancestors.append(parent)
            parent = parent.parent
        return tuple(reversed(ancestors))

    @classmethod
    def load(cls, data: Any) -> Self:
        new = fromjson(data)
        assert isinstance(new, cls)
        return new

    @classmethod
    def loads(cls, data: str) -> Self:
        return cls.load(json.loads(data))

    def children(self) -> tuple[Node, ...]:
        return self._cached_children()

    def children_list(self) -> list[Node]:
        return list(self._cached_children())

    def _cached_children(self) -> tuple[Node, ...]:
        def dfs(obj: Any) -> Iterable[Node]:
            match obj:
                case Node() as node:
                    node._parent_ref = weakref.ref(self)
                    yield node
                case Mapping() as mapping:
                    for name, value in mapping.items():
                        if name.startswith('_'):
                            continue
                        if value is None:
                            continue
                        yield from dfs(value)
                case bytes() | str():
                    pass
                case Iterable() as seq:
                    for item in seq:
                        yield from dfs(item)
                case _:
                    pass

        if self not in _children_cache:
            _children_cache[self] = tuple(dfs(self.__pub__()))
        return _children_cache[self]



from __future__ import annotations

from typing import Any

from ..contexts import Ctx
from .closure import PositiveJoin


class LeftJoin(PositiveJoin):
    JOINOP = '<'

    def _parse(self, ctx: Ctx) -> Any:
        return ctx.left_join(self.exp._parse, self.sep._parse)


class RightJoin(PositiveJoin):
    JOINOP = '>'

    def _parse(self, ctx: Ctx) -> Any:
        return ctx.right_join(self.exp._parse, self.sep._parse)



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import tatsu


def test_mixed_return():
    grammar = r"""
        start: ('a' b='b') 'c' [d='d']
    """

    model = tatsu.compile(grammar)
    value = model.parse('a b c')
    assert value == {'b': 'b', 'd': None}



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import re
import warnings
from dataclasses import dataclass
from typing import Any, Self, override

from .input import NullText
from .input.cursor import Text
from .util.configs import Config
from .util.heart import Heart
from .util.regextools import cached_re_compile
from .util.undefined import Undefined, UndefinedType
from .util.unicode_characters import C_DERIVE


DEFAULT_PERLINEMEMOS = 8
DEFAULT_HEART_BPS = 2


@dataclass
class ParserConfig(Config):
    name: str | None = 'Test'
    source: str = ''

    start: str | None = None

    semantics: Any = None

    comment_recovery: bool = False

    memoization: bool = True
    perlinememos: float = DEFAULT_PERLINEMEMOS
    memoize_lookaheads: bool | None = None
    prune_memos_on_cut: bool = True

    colorize: bool = True
    trace: bool = False
    trace_filename: str = ''
    trace_length: int = 72
    trace_separator: str = C_DERIVE

    # parser directives
    grammar: str | None = None
    left_recursion: bool = True

    comments: str | None = None
    eol_comments: str | None = None
    keywords: tuple[str, ...] | None = None

    ignorecase: bool | None = None
    namechars: str | None = None
    nameguard: bool | None = None  # implied by namechars
    whitespace: str | UndefinedType | None = Undefined
    parseinfo: bool = False
    heart: Heart | None = None
    heart_bps: float = DEFAULT_HEART_BPS

    # WARNING: DEPRECATED: some old projects use these
    owner: Any = None
    extra: Any = None
    start_rule: str | None = None
    rule_name: str | None = None
    comments_re: re.Pattern | str | None = None
    eol_comments_re: re.Pattern | str | None = None
    tokenizercls: type[Text] = NullText
    filename: str | None = None
    memo_cache_size: int | None = None

    def __post_init__(self):  # pylint: disable=W0235
        if self.ignorecase and self.keywords:
            self.keywords = tuple({k.upper() for k in self.keywords})
        super().__post_init__()

        if not self.memoization:
            self.left_recursion = False

        if self.namechars:
            self.nameguard = True

        if isinstance(self.semantics, type):
            raise TypeError(
                f'semantics must be an object instance or None, not class {self.semantics!r}',
            )

        self._check_deprecations()
        self._compile_comments()

    def _check_deprecations(self):
        if self.comments_re:
            warnings.warn(
                'ParserConfig.comments_re is deprecated: use `comments`',
                DeprecationWarning,
                stacklevel=3,
            )
            if not self.comments:
                self.comments = str(self.comments_re)
            del self.comments_re

        if self.eol_comments_re:
            warnings.warn(
                'ParserConfig.eol_comments_re is deprecated: use `eol_comments`',
                DeprecationWarning,
                stacklevel=3,
            )
            if not self.eol_comments:
                self.eol_comments = str(self.eol_comments_re)
            del self.eol_comments_re

        if self.memoize_lookaheads is not None:
            warnings.warn(
                'ParserConfig.memoize_lookaheads is deprecated and has no effect',
                DeprecationWarning,
                stacklevel=3,
            )
            del self.memoize_lookaheads

        if self.memo_cache_size is not None:
            warnings.warn(
                'ParserConfig.memo_cache_size is deprecated and has no effect',
                DeprecationWarning,
                stacklevel=3,
            )
            del self.memo_cache_size

        if not issubclass(self.tokenizercls, NullText):
            warnings.warn(
                'ParserConfig.tokenizercls is deprecated and has no effect',
                DeprecationWarning,
                stacklevel=3,
            )
            del self.tokenizercls

        if self.filename is not None:
            self.source = self.filename
            warnings.warn(
                'ParserConfig.filename is deprecated. Use .source instead',
                DeprecationWarning,
                stacklevel=3,
            )
            del self.filename

    def _compile_comments(self):
        if self.comments and isinstance(self.comments, str):
            cached_re_compile(self.comments)
        if self.eol_comments and not isinstance(self.eol_comments, re.Pattern):
            cached_re_compile(self.eol_comments)
        if self.whitespace and not isinstance(self.whitespace, re.Pattern):
            cached_re_compile(self.whitespace)

        for name in ('start_rule', 'rule_name'):
            if (value := getattr(self, name, None)) is None:
                continue
            warnings.warn(
                message=(
                    f'\nSetting {name}={value!r} is deprecated.'
                    f' Use start={value!r} for the name of the start rule.'
                ),
                category=DeprecationWarning,
                stacklevel=2,
            )
            break
        self.start = self.effective_start_rule_name()

    def effective_start_rule_name(self):
        # NOTE: there are legacy reasons for this mess
        return (
            self.start
            or getattr(self, 'start_rule', None)
            or getattr(self, 'rule_name', None)
        )

    @override
    def override(self, /, hard: bool = False, **settings: Any) -> Self:
        result = super().override(hard=hard, **settings)
        assert isinstance(result, ParserConfig)
        if 'grammar' in settings:
            result.name = result.grammar
        self._check_deprecations()
        self._compile_comments()
        return result



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from .ast import AST
from .context import ParseContext
from .ctx import CanParse, Ctx, Func
from .decorator import isname, leftrec, name, nomemo, rule, tatsumasu
from .infos import RuleInfo
from .state import _AT_, ParseState, ParseStateStack


__all__ = [
    'AST',
    'ParseContext',
    'RuleInfo',
    'CanParse',
    'Ctx',
    'Func',
    'isname',
    'name',
    'leftrec',
    'nomemo',
    'rule',
    'tatsumasu',
    '_AT_',
    'ParseState',
    'ParseStateStack',
]



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

import itertools
import string
import types
from collections.abc import Callable, Iterator
from typing import Any

from .. import peg as g
from ..config import ParserConfig
from ..contexts.ctx import Ctx
from ..exceptions import CodegenError
from ..objectmodel import Node
from ..util import Undefined, regexpp, safe_name, typename
from ..util.indent import IndentPrintMixin
from ..walkers import NodeWalker
from .boilerplt import FOOTER, HEADER, IMPORTS, PARSER_BODY


GREEKTOME = "αβδεζηθικλμνξοπρστυφχψωΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩ"
ANON = '_'


def pythongen(model: Node, parser_name: str = '') -> str:
    if isinstance(model, g.Model):
        model = model.optimized()
    generator = PythonParserGenerator(parser_name=parser_name)
    generator.walk(model)
    return generator.printed_text()


def textinputgen(model: g.Grammar, basename: str) -> str:
    generator = PythonParserGenerator(parser_name=basename)
    generator.gen_buffering(model, basename)
    return generator.printed_text()


def keywordsgen(model: g.Grammar) -> str:
    generator = PythonParserGenerator()
    generator.gen_keywords(model)
    return generator.printed_text()


class PythonParserGenerator(IndentPrintMixin, NodeWalker):
    def __init__(self, parser_name: str = ''):
        super().__init__()
        self.parser_name = parser_name
        self._block_counter: Iterator[int] = itertools.count()
        self._choice_number: int = 0
        self.ctx_stack: list[str] = ['ctx']

    def new_choice_number(self) -> int:
        n = self._choice_number
        self._choice_number += 1
        return n

    def prev_choice_number(self):
        self._choice_number -= 1

    @property
    def blockn(self) -> int:
        return next(self._block_counter)

    @property
    def loopn(self) -> str:
        n = self.blockn
        # a = GREEKTOME[n]
        a = string.ascii_letters[n]
        return f'cl{a}' if n > 0 else 'cl'

    def push_ctx(self, ctx: str):
        self.ctx_stack.append(ctx)

    def pop_ctx(self):
        self.ctx_stack.pop()

    @property
    def ctx(self):
        return self.ctx_stack[-1]

    def reset_counters(self):
        self._block_counter = itertools.count()
        self._choice_number = 0

    def walk_default(self, node: Any) -> Any:
        raise RuntimeError(f'Unknown node type: {typename(node)}')

    def walk_NIL(self, _node: g.NIL) -> Any:
        return ""

    def walk_Grammar(self, grammar: g.Grammar):
        basename = self.parser_name or grammar.name
        self.print(HEADER)
        self.print()
        self.print(IMPORTS)
        self.print()

        self.gen_keywords(grammar)
        self.gen_buffering(grammar, basename)
        self._gen_parsing(grammar, basename)

        self.print(FOOTER(name=basename))

    def walk_Rule(self, rule: g.Rule):
        def param_repr(p):
            if isinstance(p, int | float):
                return str(p)
            else:
                return repr(p.split('::')[0])

        self.reset_counters()
        params = kwparams = ''
        if rule.params:
            params = ', '.join(param_repr(p) for p in rule.params)
        if rule.kwparams:
            assert isinstance(rule.kwparams, dict)
            kwparams = ', '.join(
                f'{k}={param_repr(v)}' for k, v in rule.kwparams.items()
            )

        if params and kwparams:
            params = params + ', ' + kwparams
        elif kwparams:
            params = kwparams
        if params:
            params = f'({params})'

        # note: remove the leftrec decorator and lieave it to the analyzer
        islrec = '\n@tatsu.leftrec' if rule.is_lrec else ''
        nomemo = '\n@tatsu.nomemo' if not rule.memoizable else ''

        isname = '\n@tatsu.name' if rule.is_name else ''
        istokn = '\n@tatsu.token' if rule.is_tokn else ''

        name = safe_name(rule.name)
        self.print(f"""
                @tatsu.rule{params}\
                {islrec}\
                {nomemo}\
                {isname}\
                {istokn}\
                \ndef {name}(self, {self.ctx_stack[0]}: Ctx) -> Any:
            """)
        with self.indent():
            self.print(self.walk(rule.exp))

    def walk_BasedRule(self, rule: g.BasedRule):
        self.walk_Rule(rule)

    def walk_Call(self, call: g.Call):
        name = safe_name(call.name)
        self.print(f'self.{name}({self.ctx})')

    def walk_RuleInclude(self, include: g.RuleInclude):
        assert include.exp is not None
        self.walk(include.exp)

    def walk_Void(self, _void: g.Void):
        self.print(f'{self.ctx}.void()')

    def walk_Dot(self, _dot: g.Dot):
        self.print(f'{self.ctx}.dot()')

    def walk_Fail(self, _fail: g.Fail):
        self.print(f'{self.ctx}.fail()')

    def walk_Cut(self, _cut: g.Cut):
        self.print(f'{self.ctx}.cut()')

    def walk_Comment(self, comment: g.Comment):
        lines = '\n'.join(f'# {c!s}' for c in comment.comment.splitlines())
        self.print(f'\n{lines}\n')

    def walk_EOLComment(self, comment: g.EOLComment):
        self.walk_Comment(comment)

    def walk_EOF(self, _eof: g.EOF):
        self.print(f'{self.ctx}.eofcheck()')

    def walk_EOL(self, _eol: g.EOL):
        self.print(f'{self.ctx}.eolcheck()')

    def walk_Group(self, group: g.Group):
        self._gen_decor(Ctx.group, exp=group.exp)

    def walk_SkipGroup(self, skip: g.SkipGroup):
        self._gen_decor(Ctx.skipgroup, exp=skip.exp)

    def walk_Token(self, token: g.Token):
        self.print(f'{self.ctx}.token({token.token!r})')

    def walk_Constant(self, constant: g.Constant):
        self.print(f'{self.ctx}.constant({constant.literal!r})')

    def walk_Alert(self, alert: g.Alert):
        self.print(f'{self.ctx}.alert({alert.literal!r}, {alert.level})')

    def walk_Pattern(self, pattern: g.Pattern):
        self.print(f'{self.ctx}.pattern({regexpp(pattern.pattern)})')

    def walk_Lookahead(self, lookahead: g.Lookahead):
        self._gen_decor(Ctx.if_, exp=lookahead.exp)

    def walk_NegativeLookahead(self, lookahead: g.NegativeLookahead):
        self._gen_decor(Ctx.ifnot_, exp=lookahead.exp)

    def walk_Sequence(self, seq: g.Sequence):
        self._gen_defines_declaration(seq)
        self.walk(seq.sequence)

    def walk_Choice(self, choice: g.Choice):
        n = self.new_choice_number()
        var = GREEKTOME[n]
        outerctx = self.ctx
        # self.push_ctx(f'ctx{a}')
        try:
            self._gen_decor(Ctx.choice, ctx=outerctx, var=f'{var}')
            with self.indent():
                elements = choice.lookaheadlist
                self.pfold(f'{var}.expecting', tuple(elements))
                self.print()

                for opt in choice.options:
                    exp: g.Model = opt
                    if isinstance(exp, g.Option):
                        exp = exp.exp
                    self._gen_anon_block(exp, ctx=self.ctx, decor=f'{var}.option')
        finally:
            # self.pop_ctx()
            self.prev_choice_number()

    def walk_Option(self, _option: g.Option):
        pass  # handled by walk_Choice

    def walk_Optional(self, optional: g.Optional):
        self._gen_decor(Ctx.optional, exp=optional.exp)

    def walk_EmptyClosure(self, _closure: g.EmptyClosure):
        self.print(f'{self.ctx}.empty()')

    def walk_Closure(self, closure: g.Closure):
        self._gen_decor(Ctx.loopopt, exp=closure.exp, var=f'{self.loopn}')

    def walk_PositiveClosure(self, closure: g.PositiveClosure):
        self._gen_decor(Ctx.loopplus, exp=closure.exp, var=f'{self.loopn}')

    def walk_Join(self, join: g.Join):
        self._gen_decor(Ctx.joinopt, exp=join.exp, sep=join.sep, var=f'{self.loopn}')

    def walk_PositiveJoin(self, join: g.PositiveJoin):
        self._gen_decor(Ctx.joinplus, exp=join.exp, sep=join.sep, var=f'{self.loopn}')

    def walk_LeftJoin(self, join: g.LeftJoin):
        self._gen_decor(Ctx.joinleft, exp=join.exp, sep=join.sep, var=f'{self.loopn}')

    def walk_RightJoin(self, join: g.RightJoin):
        self._gen_decor(Ctx.joinright, exp=join.exp, sep=join.sep, var=f'{self.loopn}')

    def walk_Gather(self, gather: g.Gather):
        self._gen_decor(Ctx.gatheropt, exp=gather.exp, sep=gather.sep, var='g')

    def walk_PositiveGather(self, gather: g.PositiveGather):
        self._gen_decor(Ctx.gatherplus, exp=gather.exp, sep=gather.sep, var='g')

    def walk_SkipTo(self, skipto: g.SkipTo):
        self._gen_decor(Ctx.skipto, exp=skipto.exp)

    def walk_Named(self, named: g.Named):
        self._gen_decor(Ctx.nameset, exp=named.exp, arg=repr(named.name))

    def walk_NamedList(self, named: g.Named):
        self._gen_decor(Ctx.nameadd, exp=named.exp, arg=repr(named.name))

    def walk_Override(self, o: g.Override):
        self._gen_decor(Ctx.result, exp=o.exp)

    def walk_OverrideList(self, override: g.OverrideList):
        self._gen_decor(Ctx.resultadd, exp=override.exp)

    def gen_keywords(self, grammar: g.Grammar):
        keywords = [str(k) for k in grammar.keywords if k is not None]
        if not keywords:
            self.print('KEYWORDS = ()')
        else:
            self.print('KEYWORDS = (')
            with self.indent():
                keywords_str = '\n'.join(f'    {k!r},' for k in sorted(keywords))
                self.print(keywords_str)
            self.print(')')

        self.print()
        self.print()

    def walk_Meta(self, meta: g.Meta):
        name = meta.pretty()[1:]
        self.print(f'{self.ctx}.match{name!s}()')

    def _gen_init(self, grammar: g.Grammar):
        assert isinstance(grammar.config, ParserConfig)
        start = grammar.config.start or grammar.rules[0].name

        whitespace = grammar.config.whitespace
        if whitespace is Undefined:  # the default value
            whitespace = None
        elif whitespace is not None:
            whitespace = regexpp(whitespace)

        name = grammar.directives.get('grammar', grammar.name)
        self.print(f'''
                config = ParserConfig.new(
                    config=config,
                    name={name!r},
                    whitespace={whitespace},
                    nameguard={grammar.config.nameguard},
                    ignorecase={grammar.config.ignorecase or False},
                    namechars={grammar.config.namechars or ""!r},
                    parseinfo={grammar.config.parseinfo},
                    comments={regexpp(grammar.config.comments)},
                    eol_comments={regexpp(grammar.config.eol_comments)},
                    keywords=KEYWORDS,
                    start={start!r},
                )
                assert isinstance(config, ParserConfig)
                config = config.override(**settings)
            ''')

    def _input_name(self, basename) -> str:
        return f'{basename}Text'

    def _legacy_input_name(self, basename) -> str:
        return f'{basename}Tokenizer'

    def _buffer_name(self, basename) -> str:
        return f'{basename}Buffer'

    def _parser_name(self, basename: str) -> str:
        return f'{basename}Parser'

    def _rules_name(self, basename: str) -> str:
        return f'{basename}Rules'

    def _gen_buffering_init(self, grammar: g.Grammar):
        with self.indent():
            self.print('def __init__(')
            with self.indent():
                self.print(
                    """
                    self,
                    text,
                    /,
                    config: ParserConfig | None = None,
                    **settings,
                    """,
                )
            self.print(') -> None:')
            with self.indent():
                self._gen_init(grammar)
                self.print()
                self.print('super().__init__(text, config=config)')
        self.print()

    def gen_buffering(self, grammar: g.Grammar, basename: str):
        self.print(f'class {self._input_name(basename)}(TextLines):')
        self._gen_buffering_init(grammar)
        self.print(
            f'{self._legacy_input_name(basename)} = {self._input_name(basename)}'
        )
        self.print()

        self.print()
        self.print(
            f'class {self._buffer_name(basename)}(Buffer):  # NOTE: backwards compatibility'
        )
        self._gen_buffering_init(grammar)
        self.print()

    def _gen_parsing(self, grammar: g.Grammar, basename: str):
        self.print(f'class {self._parser_name(basename)}(Parser):')
        with self.indent():
            self.print(
                'def __init__(self, /, config: ParserConfig | None = None, **settings):'
            )
            with self.indent():
                self.print('\n' + PARSER_BODY(rules_name=self._rules_name(basename)))
        self.print()
        self.print()

        self.print(f'class {self._rules_name(basename)}:')
        with self.indent():
            self.print(
                'def __init__(self, /, config: ParserConfig | None = None, **settings):',
            )
            with self.indent():
                self._gen_init(grammar)
                self.print('self._config = config')
                self.print()
            # self.print("""
            #     @property
            #     def config(self) -> ParserConfig:
            #         assert isinstance(self._config, ParserConfig)
            #         return self._config
            #     """)
            self.print()
            self.walk(grammar.rules)
        self.print()

    def _gen_defines_declaration(self, node: g.Model):
        sdefs = node.defines_single
        ldefs = node.defines_list

        if not (sdefs or ldefs):
            return

        sdefs_str = ', '.join(sorted(repr(d) for d in sdefs))
        ldefs_str = ', '.join(sorted(repr(d) for d in ldefs))

        definestr = f'{self.ctx}.define([{sdefs_str}], [{ldefs_str}])'
        if not ldefs or self.fitsfmt(definestr, 1):
            self.print(definestr)
        else:
            self.print(f'{self.ctx}.define(')
            with self.indent():
                self.print(f'[{sdefs_str}],')
                self.print(f'[{ldefs_str}],')
            self.print(')')

    def _gen_block(self, exp: g.Model, name='block'):
        if () in exp.lookaheadlist:
            raise CodegenError(
                f'{exp!r} may repeat empty sequence @{exp.line} {exp.lookahead!r}',
            )

        n = self.blockn
        self.print()
        self.print(f'def {name}{n}():')
        with self.indent():
            self.walk(exp)

        return n

    def _gen_anon_block(
        self,
        exp: g.Model,
        decor: str = '',
        echeck: bool = False,
        ctx: str | None = None,
    ):
        ctx = ctx or self.ctx
        if echeck and () in exp.lookaheadlist:
            raise CodegenError(
                f'{exp!r} may repeat empty sequence @{exp.line} {exp.lookahead!r}',
            )

        if decor:
            self.print(f'@{decor}')
        if ctx:
            self.print(f'def {ANON}({ctx}: Ctx) -> Any:')
        else:
            self.print(f'def {ANON}() -> Any:')
        with self.indent():
            self.walk(exp)

    def _gen_decor(
        self,
        mgr: Callable[..., Any],
        /,
        *,
        exp: g.Model | None = None,
        sep: g.Model | None = None,
        var: str = '',
        arg: str = '',
        ctx: str | None = None,
        echeck: bool = True,
    ):
        assert isinstance(mgr, types.FunctionType)
        name = mgr.__name__
        if var:
            self.print(f'with ctx.{name}({arg}) as {var}:')
        else:
            self.print(f'with ctx.{name}({arg}):')
        with self.indent():
            # NOTE only generated for Choice
            # if var and exp and (elements := exp.lookaheadlist):
            #     self.pfold(f'{var}.expecting', tuple(elements))
            #     self.print()
            if sep:
                self._gen_anon_block(sep, decor=f'{var}.sep', ctx=ctx, echeck=echeck)
                self.print()
            if exp and var:
                self._gen_anon_block(exp, decor=f'{var}.exp', ctx=ctx, echeck=echeck)
            elif exp:
                self.walk(exp)



class MultiProgress:
    # ... (thread orchestration remains identical) ...

    def _render_loop(self):
        while self._running:
            with self._lock:
                snapshot = copy.deepcopy(self._bars)

            if not snapshot:
                time.sleep(0.05)
                continue

            try:
                # Use the helper attached to your Color engine
                screen_cols, _ = Color.default().terminal_size()
            except Exception:
                screen_cols = 80

            now = time.time()
            lines = []

            for b in snapshot:
                # 1. Library calculates runtime metrics safely
                elapsed = now - b.start_time
                pct = (b.current / b.total * 100) if b.total > 0 else 0.0
                rate = b.current / elapsed if elapsed > 0 else 0
                eta = (b.total - b.current) / rate if rate > 0 else 0.0

                metrics = {"pct": pct, "elapsed": elapsed, "eta": eta}

                # 2. Collect Style definitions from the component
                styles = b.render(screen_cols, metrics)

                # 3. Calculate fixed width footprint of all plain text strings
                # This ignores ANSI code lengths because Style.__str__ or Style.value
                # yields lengths accurately depending on the policy status.
                fixed_width = sum(len(s.value) for s in styles if s.value != "{bar}")
                bar_budget = max(0, screen_cols - fixed_width)

                # 4. Construct the physical layout line
                line_chunks = []
                for style in styles:
                    if style.value == "{bar}":
                        # Draw the flex component inline using calculated budget
                        fill_len = int(bar_budget * (pct / 100))
                        empty_len = bar_budget - fill_len
                        bar_str = f"[{'█' * fill_len}{'-' * empty_len}]"

                        # Apply parent style configuration to the generated bar block
                        line_chunks.append(style(bar_str))
                    else:
                        # Render the pre-configured styled string block
                        line_chunks.append(str(style))

                lines.append("".join(line_chunks))

            sys.stdout.write("\n".join(lines))
            sys.stdout.flush()

            time.sleep(0.05)  # 20 FPS
            sys.stdout.write(f"\033[{len(snapshot)}A")



# Copyright (c) 2017-2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from ..ctx import Ctx, Func
from ._base import ContextBase


class ExpContext(ContextBase):
    def __init__(self, ctx: Ctx):
        super().__init__(ctx)
        self._exp: Func | None = None

    @property
    def func(self) -> Func:
        if self._exp is None:
            raise RuntimeError('.exp not set')
        return self._exp

    def exp(self, func: Func) -> Func:
        if self._exp is not None:
            raise RuntimeError('.exp already set')
        self._exp = func
        return func



import time

start = time.monotonic_ns()
time.sleep(0.1)
end = time.monotonic_ns()

diff = end - start
print(f"Raw diff for 0.1s sleep: {diff:,} ns")
print(f"Expected around:          100,000,000 ns")



# Copyright (c) 2026 Juancarlo Añez (apalala@gmail.com)
# SPDX-License-Identifier: BSD-4-Clause
from __future__ import annotations

from tatsu.treez import Tree


def test_empty():
    t = Tree("root")
    assert t.render() == "root"


def test_one_child():
    t = Tree("root")
    t.add("child")
    assert t.render() == "root\n└── child"


def test_two_children():
    t = Tree("root")
    t.add("a")
    t.add("b")
    assert t.render() == "root\n├── a\n└── b"


def test_nested():
    t = Tree("root")
    a = t.add("a")
    a.add("x")
    t.add("b")
    assert t.render() == "root\n├── a\n│   └── x\n└── b"


def test_deep_nesting():
    t = Tree("r")
    c1 = t.add("c1")
    c2 = c1.add("c2")
    c2.add("c3")
    assert t.render() == "r\n└── c1\n    └── c2\n        └── c3"


def test_add_returns_child():
    t = Tree("root")
    child = t.add("child")
    assert child.label == "child"
    assert child in t.children


def test_str_delegates_to_render():
    t = Tree("x")
    t.add("y")
    assert str(t) == t.render()


def test_style_is_accepted():
    t = Tree("root")
    t.add("child")
    assert t.render() == "root\n└── child"


def test_no_style_default():
    t = Tree("root")
    child = t.add("child")
    assert child.label == "child"


def test_children_list():
    t = Tree("root")
    a = t.add("a")
    b = t.add("b")
    assert t.children == [a, b]


def test_three_children():
    t = Tree("root")
    t.add("a")
    t.add("b")
    t.add("c")
    expected = "root\n├── a\n├── b\n└── c"
    assert t.render() == expected


def test_single_branch_multi_level():
    t = Tree("x")
    t.add("y").add("z")
    assert t.render() == "x\n└── y\n    └── z"
